# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import stacks
from tobiko.shell import sh
from tobiko.shell import ssh


class ProcessFixture(tobiko.SharedFixture):

    temp_filename: str
    ssh_client: ssh.SSHClientType = None
    process: sh.ShellProcessFixture

    def setup_fixture(self):
        self.temp_filename = sh.execute(
            'mktemp', ssh_client=self.ssh_client).stdout.strip()
        self.addCleanup(sh.execute,
                        f"rm -f '{self.temp_filename}'",
                        ssh_client=self.ssh_client)
        self.process = sh.process(f"tail -f '{self.temp_filename}'",
                                  ssh_client=self.ssh_client)
        self.process.execute()
        self.addCleanup(self.process.kill)


class ProcessTest(testtools.TestCase):

    fixture = tobiko.required_fixture(ProcessFixture)

    def test_stdout(self):
        fixture = self.fixture
        sh.execute(f"echo some text > '{fixture.temp_filename}'",
                   ssh_client=fixture.ssh_client)
        line = self.fixture.process.stdout.readline()
        self.assertEqual(b'some text\n', line)


class LocalProcessFixture(ProcessFixture):

    ssh_client: ssh.SSHClientType = False


class LocalProcessTest(ProcessTest):

    fixture = tobiko.required_fixture(LocalProcessFixture)


class SSHProcessFixture(ProcessFixture):

    stack = tobiko.required_fixture(
        stacks.UbuntuMinimalServerStackFixture)

    def setup_fixture(self):
        self.ssh_client = self.stack.ssh_client
        super().setup_fixture()


@keystone.skip_unless_has_keystone_credentials()
class SSHProcessTest(ProcessTest):

    fixture = tobiko.required_fixture(SSHProcessFixture)


class CirrosProcessFixture(SSHProcessFixture):

    stack = tobiko.required_fixture(
        stacks.CirrosServerStackFixture)


@keystone.skip_unless_has_keystone_credentials()
class CirrosProcessTest(ProcessTest):

    fixture = tobiko.required_fixture(CirrosProcessFixture)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import typing

import metalsmith
import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import topology
from tobiko.shell import files
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import _overcloud
from tobiko.tripleo import _rhosp
from tobiko.tripleo import _undercloud

CONF = config.CONF
LOG = log.getLogger(__name__)


class TripleoTopology(topology.OpenStackTopology):

    agent_to_service_name_mappings = {
        neutron.DHCP_AGENT: 'tripleo_neutron_dhcp',
        neutron.L3_AGENT:  'tripleo_neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'tripleo_neutron_ovs_agent',
        neutron.METADATA_AGENT: 'tripleo_neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'tripleo_ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'tripleo_ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'tripleo_ovn_controller'
    }

    agent_to_container_name_mappings = {
        neutron.DHCP_AGENT: 'neutron_dhcp',
        neutron.L3_AGENT:  'neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'neutron_ovs_agent',
        neutron.METADATA_AGENT: 'neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'ovn_controller'
    }

    has_containers = True

    config_file_mappings = {
        'ml2_conf.ini': '/var/lib/config-data/puppet-generated/neutron'
                        '/etc/neutron/plugins/ml2/ml2_conf.ini'
    }

    # TODO: add more known subgrups here
    known_subgroups: typing.List[str] = ['controller', 'compute']

    # In TripleO we need to parse log files directly
    file_digger_class = files.LogFileDigger

    # This is dict which handles mapping of the log file and systemd_unit (if
    # needed) for the OpenStack services
    # Format of this dict is like below:
    # service_name: (log_filename, systemd_unit_name)
    log_names_mappings = {
        neutron.SERVER: '/var/log/containers/neutron/server.log*',
    }

    def create_node(self, name, ssh_client, **kwargs):
        return TripleoTopologyNode(topology=self,
                                   name=name,
                                   ssh_client=ssh_client,
                                   **kwargs)

    def discover_nodes(self):
        self.discover_ssh_proxy_jump_node()
        self.discover_undercloud_nodes()
        self.discover_overcloud_nodes()

    def discover_undercloud_nodes(self):
        if _undercloud.has_undercloud():
            uc_config = _undercloud.undercloud_host_config()
            ssh_client = _undercloud.undercloud_ssh_client()
            self.add_node(address=uc_config.hostname,
                          group='undercloud',
                          ssh_client=ssh_client)

    def discover_overcloud_nodes(self):
        if _undercloud.has_undercloud():
            for instance in _overcloud.list_overcloud_nodes():
                try:
                    _overcloud.power_on_overcloud_node(instance)
                except Exception:
                    LOG.exception("Error ensuring overcloud node power "
                                  "status is on")
                host_config = _overcloud.overcloud_host_config(
                    instance=instance)
                ssh_client = _overcloud.overcloud_ssh_client(
                    instance=instance,
                    host_config=host_config)
                node = self.add_node(address=host_config.hostname,
                                     group='overcloud',
                                     ssh_client=ssh_client,
                                     overcloud_instance=instance)
                assert isinstance(node, TripleoTopologyNode)
                self.discover_overcloud_node_subgroups(node)

    def discover_overcloud_node_subgroups(self, node):
        # set of subgroups extracted from node name
        subgroups: typing.Set[str] = set()

        oc_groups_dict = CONF.tobiko.tripleo.overcloud_groups_dict
        # extract subgroups names from node name
        subgroups.update(oc_groups_dict.get(subgroup) or subgroup
                         for subgroup in node.name.split('-')
                         if is_valid_overcloud_group_name(group_name=subgroup,
                                                          node_name=node.name))

        # add all those known subgroups names that are contained in
        # the node name (controller, compute, ...)
        subgroups.update(subgroup
                         for subgroup in self.known_subgroups
                         if subgroup in node.name)

        # bind node to discovered subgroups
        if subgroups:
            for subgroup in sorted(subgroups):
                LOG.debug("Add node '%s' to subgroup '%s'", node.name,
                          subgroup)
                self.add_node(hostname=node.name, group=subgroup)
        else:
            LOG.warning("Unable to obtain any node subgroup from node "
                        "name: '%s'", node.name)
        return subgroups


class TripleoTopologyNode(topology.OpenStackTopologyNode):

    def __init__(self,
                 topology: topology.OpenStackTopology,
                 name: str,
                 ssh_client: ssh.SSHClientFixture,
                 addresses: typing.Iterable[netaddr.IPAddress],
                 hostname: str,
                 overcloud_instance: metalsmith.Instance = None,
                 rhosp_version: tobiko.Version = None):
        # pylint: disable=redefined-outer-name
        super().__init__(topology=topology,
                         name=name,
                         ssh_client=ssh_client,
                         addresses=addresses,
                         hostname=hostname)
        self._overcloud_instance = overcloud_instance
        self._rhosp_version = rhosp_version

    @property
    def overcloud_instance(self) -> typing.Optional[metalsmith.Instance]:
        return self._overcloud_instance

    @property
    def rhosp_version(self) -> tobiko.Version:
        if self._rhosp_version is None:
            self._rhosp_version = self._get_rhosp_version()
        return self._rhosp_version

    l3_agent_conf_path = (
        '/var/lib/config-data/neutron/etc/neutron/l3_agent.ini')

    def reboot_overcloud_node(self,
                              reactivate_servers=True):
        """Reboot overcloud node

        This method reboots an overcloud node and may start every Nova
        server which is not in SHUTOFF status before restarting.

        :param reactivate_servers: whether or not to re-start the servers which
            are hosted on the compute node after the reboot
        """

        running_servers: typing.List[nova.NovaServer] = []
        if reactivate_servers:
            running_servers = self.list_running_servers()
            LOG.debug(f'Servers to restart after reboot: {running_servers}')

        self.power_off_overcloud_node()
        self.power_on_overcloud_node()

        if running_servers:
            LOG.info(f'Restart servers after rebooting overcloud compute node '
                     f'{self.name}...')
            for server in running_servers:
                nova.wait_for_server_status(server=server.id,
                                            status='SHUTOFF')
                LOG.debug(f'Re-activate server {server.name} with ID '
                          f'{server.id}')
                nova.activate_server(server=server)
                LOG.debug(f'Server {server.name} with ID {server.id} has '
                          f'been reactivated')

    def list_running_servers(self) -> typing.List[nova.NovaServer]:
        running_servers = list()
        for server in nova.list_servers():
            if server.status != 'SHUTOFF':
                hypervisor_name = nova.get_server_hypervisor(server,
                                                             short=True)
                if self.name == hypervisor_name:
                    running_servers.append(server)
        return running_servers

    def power_on_overcloud_node(self):
        if self.overcloud_instance is None:
            raise TypeError(f"Node {self.name} is not and Overcloud server")
        self.ssh_client.close()
        LOG.debug(f"Ensuring overcloud node {self.name} power is on...")
        _overcloud.power_on_overcloud_node(instance=self.overcloud_instance)
        hostname = sh.get_hostname(ssh_client=self.ssh_client)
        LOG.debug(f"Overcloud node {self.name} power is on ("
                  f"hostname={hostname})")

    def power_off_overcloud_node(self):
        if self.overcloud_instance is None:
            raise TypeError(f"Node {self.name} is not and Overcloud server")
        self.ssh_client.close()
        LOG.debug(f"Ensuring overcloud node {self.name} power is off...")
        _overcloud.power_off_overcloud_node(instance=self.overcloud_instance)
        LOG.debug(f"Overcloud server node {self.name} power is off.")

    def _get_rhosp_version(self) -> tobiko.Version:
        return _rhosp.get_rhosp_version(connection=self.connection)


def is_valid_overcloud_group_name(group_name: str, node_name: str = None):
    if not group_name:
        return False
    if group_name in ['overcloud', node_name]:
        return False
    if is_number(group_name):
        return False
    return True


def is_number(text: str):
    try:
        float(text)
    except ValueError:
        return False
    else:
        return True


def setup_tripleo_topology():
    if _undercloud.has_undercloud() or _overcloud.has_overcloud():
        topology.set_default_openstack_topology_class(TripleoTopology)


def get_ip_to_nodes_dict(openstack_nodes=None):
    if not openstack_nodes:
        openstack_nodes = topology.list_openstack_nodes(group='overcloud')
    ip_to_nodes_dict = {str(node.public_ip): node.name for node in
                        openstack_nodes}
    return ip_to_nodes_dict


def str_is_not_ip(check_str):
    letters = re.compile('[A-Za-z]')
    return bool(letters.match(check_str))


def ip_to_hostname(oc_ip):
    return get_ip_to_nodes_dict()[oc_ip]


def actual_node_groups(groups):
    """return only existing node groups"""
    return set(groups).intersection(topology.list_openstack_node_groups())


def get_node(node_name):
    node_name = node_name.split('.')[0]
    return [node for node in topology.list_openstack_nodes() if
            node.name == node_name][0]


def is_composable_roles_env():
    composable_nodes = ['messaging', 'database', 'networker']
    for nodes in composable_nodes:
        if nodes in topology.list_openstack_node_groups():
            return True
    return False



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


from tobiko.shell import sh
from tobiko.tests import unit


class JoinChunksTest(unit.TobikoUnitTest):

    def test_join_chunks(self, chunks=None, expected_result=None):
        chunks = chunks or []
        actual_result = sh.join_chunks(chunks)
        self.assertIsInstance(actual_result, type(expected_result))
        if len(chunks) > 1:
            self.assertEqual(expected_result, actual_result)
        else:
            self.assertIs(expected_result, actual_result)

    def test_join_chunks_with_bytes(self):
        self.test_join_chunks([b'a', b'b', b'c'], b'abc')

    def test_join_chunks_with_one_bytes(self):
        self.test_join_chunks([b'abc'], b'abc')

    def test_join_chunks_with_bytes_and_nones(self):
        self.test_join_chunks([None, b'ab', None, b'cd'], b'abcd')

    def test_join_chunks_with_strings(self):
        self.test_join_chunks(['a', 'b', 'c'], 'abc')

    def test_join_chunks_with_one_string(self):
        self.test_join_chunks(['abc'], 'abc')

    def test_join_chunks_with_strings_and_nones(self):
        self.test_join_chunks([None, 'ab', None, 'cd'], 'abcd')

    def test_join_chunks_with_unicodes(self):
        self.test_join_chunks([u'a', u'b', u'c'], u'abc')

    def test_join_chunks_with_one_unicode(self):
        self.test_join_chunks([u'abc'], u'abc')

    def test_join_chunks_with_unicodes_and_nones(self):
        self.test_join_chunks([None, u'ab', None, u'cd'], u'abcd')



from __future__ import absolute_import

import io
import re
import time

from oslo_log import log
import pandas

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.tripleo import overcloud
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class OvercloudProcessesException(tobiko.TobikoException):
    message = "not all overcloud processes are in running state, " \
              "{process_error}"


def get_overcloud_node_processes_table(ssh_client: ssh.SSHClientType):
    """
    get processes tables from overcloud node

       returns :
[root@controller-0 ~]# ps -axw -o "%U" -o "|%p" -o "|%P" -o "|%C" -o "|%z" -o
"|%x" -o "|%c" -o "|%a" |grep -v 'ps -aux'|head
USER    |    PID|   PPID|%CPU|   VSZ|    TIME|COMMAND        |COMMAND
|overcloud_node
root    |      1|      0| 1.3|246892|01:08:57|systemd        |/usr/lib/systemd
controller-0 ...
/systemd --switched-root --system --deserialize 18
root    |      2|      0| 0.0|     0|00:00:00|kthreadd       |[kthreadd]
root    |      3|      2| 0.0|     0|00:00:00|rcu_gp         |[rcu_gp]
root    |      4|      2| 0.0|     0|00:00:00|rcu_par_gp     |[rcu_par_gp]
root    |      6|      2| 0.0|     0|00:00:00|kworker/0:0H-ev|[kworker/0:0H
-events_highpri]
root    |      8|      2| 0.0|     0|00:00:00|mm_percpu_wq   |[mm_percpu_wq]
root    |      9|      2| 0.0|     0|00:00:06|ksoftirqd/0    |[ksoftirqd/0]
root    |     10|      2| 0.0|     0|00:04:28|rcu_sched      |[rcu_sched]
root    |     11|      2| 0.0|     0|00:00:05|migration/0    |[migration/0]

    :return: dataframe of overcloud node processes dataframe
    """

    output = sh.execute(
        "ps -axw -o \"%U\" -o \"DELIM%p\" -o \"DELIM%P\" -o \"DELIM%C\" -o "
        "\"DELIM%z\" -o \"DELIM%x\" -o \"DELIM%c\" -o \"DELIM%a\" |grep -v "
        "'ps -axw' |sed 's/\"/''/g'",
        ssh_client=ssh_client).stdout
    stream = io.StringIO(output)
    table: pandas.DataFrame = pandas.read_csv(
        stream, sep='DELIM', header=None, skiprows=1)
    table.replace(to_replace=' ', value="", regex=True, inplace=True)
    table.columns = ['USER', 'PID', 'PPID', 'CPU', 'VSZ', 'TIME', 'PROCESS',
                     'PROCESS_ARGS']
    # pylint: disable=unsupported-assignment-operation
    hostname = sh.get_hostname(ssh_client=ssh_client)
    table['overcloud_node'] = hostname

    LOG.debug("Successfully got overcloud nodes processes status table")
    return table


def get_overcloud_nodes_running_process(process):
    """
    Check what nodes are running the specifies
    process: exact str of a process name as seen in ps -axw -o "%c"
    :return: list of overcloud nodes
    """
    oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_processes_table)
    oc_nodes_running_process = oc_procs_df.query('PROCESS=="{}"'.format(
        process))['overcloud_node'].unique()
    return oc_nodes_running_process


def check_if_process_running_on_overcloud(process):
    """
    Check what nodes are running the specifies
    process: exact str of a process name as seen in ps -axw -o "%c"
    :return: list of overcloud nodes
    """
    oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                          get_overcloud_node_processes_table)
    if not oc_procs_df.query('PROCESS=="{}"'.format(process)).empty:
        return True
    else:
        return False


class OvercloudProcessesStatus(object):
    """
    class to handle processes checks,
    checks that all of these are running in the overcloud:
    'ovsdb-server','pcsd', 'corosync', 'beam.smp', 'mysqld', 'redis-server',
    'haproxy', 'nova-conductor', 'nova-scheduler', 'neutron-server',
     'nova-compute', 'glance-api'
    """
    def __init__(self):
        self.processes_to_check = ['ovsdb-server', 'pcsd', 'corosync',
                                   'beam.smp', 'mysqld', 'redis-server',
                                   'haproxy', 'nova-conductor',
                                   'nova-scheduler', 'neutron-server',
                                   'nova-compute', 'glance-api']

        num_northd_proc = 'all' if overcloud.is_ovn_using_raft() else 1
        self.ovn_processes_to_check_per_node = [{'name': 'ovn-controller',
                                                 'node_group': 'controller',
                                                 'number': 'all'},
                                                {'name': 'ovn-controller',
                                                 'node_group': 'compute',
                                                 'number': 'all'},
                                                {'name': 'ovn-northd',
                                                 'node_group': 'controller',
                                                 'number': num_northd_proc}]

        self.oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_processes_table)

    @property
    def basic_overcloud_processes_running(self):
        """
        Checks that the oc_procs_df dataframe has all of the list procs
        :return: Bool
        """
        for attempt_number in range(600):

            try:

                for process_name in self.processes_to_check:
                    # osp16/python3 process is "neutron-server:"
                    if process_name == 'neutron-server' and \
                            self.oc_procs_df.query('PROCESS=="{}"'.format(
                            process_name)).empty:
                        process_name = 'neutron-server:'
                    # osp17 mysqld process name is mysqld_safe
                    if process_name == 'mysqld' and \
                            self.oc_procs_df.query('PROCESS=="{}"'.format(
                            process_name)).empty:
                        process_name = 'mysqld_safe'
                    # redis not deployed on osp17 by default, only if some
                    # other services such as designate and octavia are deployed
                    if (process_name == 'redis-server' and
                            not overcloud.is_redis_expected()):
                        redis_message = ("redis-server not expected on OSP 17 "
                                         "and later releases by default")
                        if self.oc_procs_df.query(
                                f'PROCESS=="{process_name}"').empty:
                            LOG.info(redis_message)
                            continue
                        else:
                            raise OvercloudProcessesException(
                                process_error=redis_message)

                    if not self.oc_procs_df.query('PROCESS=="{}"'.format(
                            process_name)).empty:
                        LOG.info("overcloud processes status checks: "
                                 "process {} is  "
                                 "in running state".format(process_name))
                        continue
                    else:
                        LOG.info("Failure : overcloud processes status checks:"
                                 "process {} is not running ".format(
                                  process_name))
                        raise OvercloudProcessesException(
                            process_error="process {} is not running ".format(
                                          process_name))
                # if all procs are running we can return true
                return True
            except OvercloudProcessesException:
                LOG.info('Retrying overcloud processes checks attempt '
                         '{} of 360'.format(attempt_number))
                time.sleep(1)
                self.oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                    get_overcloud_node_processes_table)
        # exhausted all retries
        tobiko.fail('Not all overcloud processes are running !\n')

    @property
    def ovn_overcloud_processes_validations(self):
        """
        Checks that the oc_procs_df dataframe has OVN processes running on the
        expected overcloud node or nodes
        :return: Bool
        """
        if not neutron.has_ovn():
            LOG.info("Networking OVN not configured")
            return True

        for process_dict in self.ovn_processes_to_check_per_node:
            if not self.oc_procs_df.query('PROCESS=="{}"'.format(
                    process_dict['name'])).empty:
                LOG.info("overcloud processes status checks: "
                         f"process {process_dict['name']} is  "
                         "in running state")

                ovn_proc_filtered_df = self.oc_procs_df.query(
                    'PROCESS=="{}"'.format(process_dict['name']))

                if (process_dict['node_group'] not in
                        topology.list_openstack_node_groups()):
                    LOG.debug(f"{process_dict['node_group']} is not "
                              "a node group part of this Openstack cloud")
                    continue
                node_list = [node.name
                             for node in
                             topology.list_openstack_nodes(
                                group=process_dict['node_group'])]
                node_names_re = re.compile(r'|'.join(node_list))
                node_filter = (ovn_proc_filtered_df.overcloud_node.
                               str.match(node_names_re))
                # obtain the processes running on a specific type of nodes
                ovn_proc_filtered_per_node_df = \
                    ovn_proc_filtered_df[node_filter]
                if type(process_dict['number']) == int:
                    assert process_dict['number'] == \
                        len(ovn_proc_filtered_per_node_df), (
                        "Unexpected number"
                        f" of processes {process_dict['name']} running on "
                        f"{process_dict['node_group']} nodes")
                elif process_dict['number'] == 'all':
                    num_nodes = len(node_list)
                    assert num_nodes == len(ovn_proc_filtered_per_node_df), (
                        "Unexpected number of processes "
                        f"{process_dict['name']} running on "
                        f"{process_dict['node_group']} nodes")
                else:
                    raise RuntimeError("Unexpected value:"
                                       f"{process_dict['node_group']}")
                # process successfully validated
                LOG.debug(f"{process_dict['name']} successfully validated on "
                          f"{process_dict['node_group']} nodes")

        # if all procs are running we can return true
        return True



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa

import tobiko
from tobiko.tests import unit


class Operation(tobiko.Operation):
    executed = False

    def run_operation(self):
        self.executed = True


class OperationForDecorator(Operation):
    pass


class OperationTest(unit.TobikoUnitTest):

    operation = tobiko.runs_operation(Operation)
    expect_after_operations = set()  # type: typing.Set[str]
    expect_executed_operations = set()  # type: typing.Set[str]

    def expect_is_after(self, operation):
        name = tobiko.get_operation_name(operation)
        return name in self.expect_after_operations

    def expect_is_executed(self, operation):
        name = tobiko.get_operation_name(operation)
        return name in self.expect_executed_operations

    def test_operation_config(self):
        config = tobiko.operation_config()
        self.assertFalse(config.run_operations)
        self.assertEqual(set(), config.after_operations)

    def test_operation(self, operation=None):
        operation = operation or self.operation
        expect_is_after = self.expect_is_after(operation)
        self.assertIs(not expect_is_after, operation.is_before)
        self.assertIs(expect_is_after, operation.is_after)
        expect_is_executed = self.expect_is_executed(operation)
        self.assertIs(expect_is_executed, operation.executed)

    @tobiko.with_operation(OperationForDecorator)
    def test_with_operation(self):
        operation = tobiko.get_operation(OperationForDecorator)
        self.test_operation(operation)

    @tobiko.before_operation(OperationForDecorator)
    def test_before_operation(self):
        operation = tobiko.get_operation(OperationForDecorator)
        self.test_operation(operation)

    @tobiko.after_operation(OperationForDecorator)
    def test_after_operation(self):
        operation = tobiko.get_operation(OperationForDecorator)
        if self.expect_is_after(operation):
            self.test_operation(operation)
        else:
            self.fail('Test method not skipped')

    @operation.with_operation
    def test_with_operation_method(self):
        self.test_operation()

    @operation.before_operation
    def test_before_operation_method(self):
        self.test_operation()

    @operation.after_operation
    def test_after_operation_method(self):
        if self.expect_is_after(self.operation):
            self.test_operation()
        else:
            self.fail('Test method not skipped')


class DontRunOperationsTest(OperationTest):

    patch_environ = {
        'TOBIKO_RUN_OPERATIONS': 'false'
    }


class DoRunOperationsTest(OperationTest):

    patch_environ = {
        'TOBIKO_RUN_OPERATIONS': 'true'
    }

    expect_run_operations = True
    expect_after_operations = {
        tobiko.get_operation_name(Operation),
        tobiko.get_operation_name(OperationForDecorator)}
    expect_executed_operations = expect_after_operations

    def test_operation_config(self):
        config = tobiko.operation_config()
        self.assertTrue(config.run_operations)
        self.assertEqual(set(), config.after_operations)


class AfterOperationsTest(OperationTest):

    patch_environ = {
        'TOBIKO_AFTER_OPERATIONS': tobiko.get_fixture_name(Operation)
    }

    expect_after_operations = {tobiko.get_operation_name(Operation)}

    def test_operation_config(self):
        config = tobiko.operation_config()
        self.assertFalse(config.run_operations)
        self.assertEqual(self.expect_after_operations,
                         config.after_operations)


class RunOperationsMixinTest(tobiko.RunsOperations, unit.TobikoUnitTest):

    patch_environ = {
        'TOBIKO_RUN_OPERATIONS': 'true'
    }

    operation = tobiko.runs_operation(Operation)

    def test_operation(self):
        self.assertTrue(self.operation.is_after)
        self.assertTrue(self.operation.executed)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.openstack import heat
from tobiko.shiftstack import _keystone


def shiftstack_heat_client(obj: heat.HeatClientType) -> heat.HeatClient:
    if obj is None:
        return get_shiftstack_heat_client()
    else:
        return tobiko.check_valid_type(obj, heat.HeatClient)


def get_shiftstack_heat_client() -> heat.HeatClient:
    session = _keystone.shiftstack_keystone_session()
    return heat.get_heat_client(session=session)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.octavia import _amphora
from tobiko.openstack.octavia import _client
from tobiko.openstack.octavia import _constants
from tobiko.openstack.octavia import _exceptions
from tobiko.openstack.octavia import _load_balancer
from tobiko.openstack.octavia import _validators
from tobiko.openstack.octavia import _waiters

AmphoraIdType = _amphora.AmphoraIdType
AmphoraType = _amphora.AmphoraType
get_amphora_id = _amphora.get_amphora_id
get_amphora = _amphora.get_amphora
get_amphora_compute_node = _amphora.get_amphora_compute_node
get_master_amphora = _amphora.get_master_amphora
list_amphorae = _amphora.list_amphorae
get_amphora_stats = _amphora.get_amphora_stats
run_command_on_amphora = _amphora.run_command_on_amphora

OCTAVIA_CLIENT_CLASSSES = _client.OCTAVIA_CLIENT_CLASSSES
get_octavia_client = _client.get_octavia_client
octavia_client = _client.octavia_client
OctaviaClientFixture = _client.OctaviaClientFixture
OctaviaClientType = _client.OctaviaClientType
get_member = _client.get_member
list_members = _client.list_members

LoadBalancerIdType = _load_balancer.LoadBalancerIdType
LoadBalancerType = _load_balancer.LoadBalancerType
get_load_balancer = _load_balancer.get_load_balancer
get_load_balancer_id = _load_balancer.get_load_balancer_id

# Waiters
wait_for_status = _waiters.wait_for_status

# Validators
check_members_balanced = _validators.check_members_balanced

# Exceptions
RequestException = _exceptions.RequestException
TimeoutException = _exceptions.TimeoutException
OctaviaClientException = _exceptions.OctaviaClientException
RoundRobinException = _exceptions.RoundRobinException
TrafficTimeoutError = _exceptions.TrafficTimeoutError
AmphoraMgmtPortNotFound = _exceptions.AmphoraMgmtPortNotFound

# Constants
PROVISIONING_STATUS = _constants.PROVISIONING_STATUS
ACTIVE = _constants.ACTIVE
ERROR = _constants.ERROR
PENDING_UPDATE = _constants.PENDING_UPDATE
WORKER_SERVICE = _constants.WORKER_SERVICE
HOUSEKEEPING_SERVICE = _constants.HOUSEKEEPING_SERVICE
HM_SERVICE = _constants.HM_SERVICE
API_SERVICE = _constants.API_SERVICE
WORKER_CONTAINER = _constants.WORKER_CONTAINER
HOUSEKEEPING_CONTAINER = _constants.HOUSEKEEPING_CONTAINER
HM_CONTAINER = _constants.HM_CONTAINER
API_CONTAINER = _constants.API_CONTAINER
OCTAVIA_SERVICES = [WORKER_SERVICE, HOUSEKEEPING_SERVICE, HM_SERVICE,
                    API_SERVICE]
OCTAVIA_CONTAINERS = [WORKER_CONTAINER, HOUSEKEEPING_CONTAINER, HM_CONTAINER,
                      API_CONTAINER]



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

from tobiko.shell import sh
from tobiko.tests.faults.containers import container_ops


LOG = log.getLogger(__name__)


@container_ops.skip_unless_has_docker
class ConfigurationFilesTest(testtools.TestCase):

    def check_config(self, node, containers, file_list, service):
        """Verify that files on the node are equal to the file in container

        It calculates MD5 sum of the configuration file in the changed
        directory if it exists there or in original directory otherwise.
        It also follows any links if necessary.
        """
        original_dir = f'/var/lib/config-data/{service}'
        changed_dir = f'/var/lib/config-data/puppet-generated/{service}'
        verified = True
        skip_container_list = ['ovn-dbs-bundle']
        for fname in file_list:
            flink = sh.execute(
                    f'sudo readlink {changed_dir}{fname} || '
                    f'sudo readlink {original_dir}{fname}',
                    ssh_client=node.ssh_client,
                    expect_exit_status=None).stdout.strip() or fname
            md5_output = sh.execute(
                    f'sudo md5sum {changed_dir}{flink} || '
                    f'sudo md5sum {changed_dir}{fname} || '
                    f'sudo md5sum {original_dir}{flink} ||'
                    f'sudo md5sum {original_dir}{fname}',
                    ssh_client=node.ssh_client,
                    expect_exit_status=None).stdout.strip().split(' ')
            if md5_output:
                md5 = md5_output[0]
                LOG.debug(f'{md5_output[-1]} on {node.name} has {md5} MD5')
            else:
                md5 = ''
                verified = False
                LOG.error(f'{node.name}: {fname} does not exist in '
                          f'{original_dir} or {changed_dir}')
            for container in containers:
                skip_this_container = False
                for skip_container in skip_container_list:
                    if skip_container in container:
                        skip_this_container = True
                if skip_this_container:
                    continue
                # 'docker' is used here in order to be compatible with old OSP
                # versions. On versions with podman, 'docker' command is
                # linked to 'podman'
                result = sh.execute(
                        f"sudo docker exec -u root {container} md5sum "
                        f"{fname} | awk '{{print $1}}'",
                        ssh_client=node.ssh_client)
                container_md5 = result.stdout.strip()
                LOG.debug(f'{fname} in {container} container '
                          f'has {container_md5} md5 hash')
                if container_md5 != md5:
                    LOG.error(f'incorrect md5 for {fname}. '
                              f'{node.name}: {md5}, {container} container: '
                              f'{container_md5}')
                    verified = False
        return verified

    def test_neutron_config_files(self):
        groups = ['controller', 'compute', 'networker']
        neutron_nodes = container_ops.get_nodes_for_groups(groups)
        for node in neutron_nodes:
            containers_neutron = (container_ops.
                                  get_node_neutron_containers(node))
            config_neutron = container_ops.get_node_neutron_config_files(node)
            self.assertTrue(
                    self.check_config(node, containers_neutron,
                                      config_neutron, 'neutron'))
            containers_ovn = container_ops.get_node_ovn_containers(node)
            config_ovn = container_ops.get_node_ovn_config_files()
            self.assertTrue(
                    self.check_config(node, containers_ovn,
                                      config_ovn, 'ovn_controller'))



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.openstack.nova import _client


class HypervisorListFixture(tobiko.SharedFixture):

    client = None
    hypervisors = None

    def setup_fixture(self):
        self.setup_client()
        self.get_hypervisors()

    def setup_client(self):
        self.client = _client.get_nova_client()

    def get_hypervisors(self):
        self.hypervisors = _client.list_hypervisors(client=self.client)


def get_hypervisors(**params):
    hypervisors = tobiko.setup_fixture(HypervisorListFixture).hypervisors
    return tobiko.select(hypervisors).with_attributes(**params)


def missing_hypervisors(count=1, **params):
    agents = get_hypervisors(**params)
    return max(0, count - len(agents))


def skip_if_missing_hypervisors(count=1, **params):
    message = "missing {return_value!r} hypervisor(s)"
    if params:
        message += " with {!s}".format(
            ', '.join("{!s}={!r}".format(k, v) for k, v in params.items()))
    return tobiko.skip_if(message, missing_hypervisors, count=count,
                          **params)


def list_servers_hypervisors(servers: typing.Iterable[_client.ServerType],
                             client: _client.NovaClientType = None) \
        -> tobiko.Selection[str]:
    hypervisors = tobiko.Selection[str]()
    client = _client.nova_client(client)
    for server in (servers or list()):
        server = _client.get_server(server, client=client)
        hypervisors.append(get_server_hypervisor(server))
    return hypervisors


def get_server_hypervisor(server: _client.ServerType,
                          client: _client.NovaClientType = None,
                          short=False) -> str:
    if isinstance(server, str):
        server = _client.get_server(server, client=client)
    hypervisor = getattr(server, 'OS-EXT-SRV-ATTR:hypervisor_hostname')
    if short:
        hypervisor = hypervisor.split('.', 1)[0]
    return hypervisor



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from datetime import datetime
import os
import subprocess

from oslo_log import log
from py.xml import html  # pylint: disable=no-name-in-module,import-error
import pytest
from pytest_html import plugin as html_plugin

import tobiko

LOG = log.getLogger(__name__)


def normalize_path(path):
    return os.path.realpath(os.path.expanduser(path))


TOP_DIR = normalize_path(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))))


# Output dirs
REPORT_DIR = os.path.realpath(os.path.expanduser(
    os.environ.get('TOBIKO_TEST_PATH') or
    os.environ.get('TOX_REPORT_DIR') or
    os.getcwd()))

REPORT_NAME = (
    os.environ.get('TOBIKO_REPORT_NAME') or
    os.environ.get('TOX_REPORT_NAME') or
    'tobiko_results')

REPORT_PREFIX = os.path.join(REPORT_DIR, REPORT_NAME)

REPORT_HTML = (
    os.environ.get('TOBIKO_REPORT_HTML') or
    os.environ.get('TOX_REPORT_HTML') or
    REPORT_PREFIX + '.html')


@pytest.hookimpl
def pytest_configure(config):
    configure_metadata(config)
    configure_caplog(config)
    configure_timeout(config)
    configure_junitxml(config)
    configure_html(config)


def configure_metadata(config):
    # pylint: disable=protected-access
    from tobiko import version
    config._metadata["Tobiko Version"] = version.release
    git_commit = subprocess.check_output(
        ['git', 'log', '-n', '1'],
        universal_newlines=True).replace('\n', '<br>')
    config._metadata["Tobiko Git Commit"] = git_commit
    git_release = subprocess.check_output(
        ['git', 'describe', '--tags'],
        universal_newlines=True).replace('\n', '<br>')
    config._metadata["Tobiko Git Release"] = git_release


def configure_caplog(config):
    tobiko_config = tobiko.tobiko_config()

    if tobiko_config.logging.capture_log:
        if tobiko_config.debug:
            level = 'DEBUG'
        else:
            level = 'INFO'
    else:
        level = 'FATAL'
    for key in ['log_level',
                'log_file_level',
                'log_cli_level']:
        set_default_inicfg(config, key, level)

    line_format: str = tobiko_config.logging.line_format
    if line_format:
        # instance and color are not supported by pytest
        line_format = line_format.replace('%(instance)s', '')
        line_format = line_format.replace('%(color)s', '')
        if line_format:
            for key in ['log_format',
                        'log_file_format',
                        'log_cli_format']:
                set_default_inicfg(config, key, line_format)

    date_format = tobiko_config.logging.date_format
    if date_format:
        for key in ['log_date_format',
                    'log_file_date_format',
                    'log_cli_date_format']:
            set_default_inicfg(config, key, date_format)


def configure_junitxml(config):
    config.inicfg['junit_suite_name'] = REPORT_NAME


def configure_html(config):
    if config.pluginmanager.hasplugin('html'):
        htmlpath = config.getoption('htmlpath')
        if htmlpath is None:
            config.option.htmlpath = REPORT_HTML
            htmlpath = config.getoption('htmlpath')
        assert htmlpath is not None

        html_plugin.HTMLReport = HTMLReport


class HTMLReport(html_plugin.HTMLReport):

    session = None
    last_html_generation: tobiko.Seconds = None

    def pytest_sessionstart(self, session):
        super().pytest_sessionstart(session)
        self.session = session

    def pytest_runtest_logreport(self, report):
        super().pytest_runtest_logreport(report)

        # Avoid report regeneration ad an interval smaller than 10 seconds
        now = tobiko.time()
        if (self.last_html_generation is not None and
                now - self.last_html_generation < 10.):
            return
        self.last_html_generation = now

        LOG.debug("Update HTML test report files...")
        temp_report = html_plugin.HTMLReport(logfile=self.logfile,
                                             config=self.config)
        # pylint: disable=attribute-defined-outside-init
        temp_report.suite_start_time = self.suite_start_time
        temp_report.reports = dict(self.reports)
        temp_report.pytest_sessionfinish(self.session)
        LOG.debug("HTML test report files updated")


def set_default_inicfg(config, key, default):
    value = config.inicfg.setdefault(key, default)
    if value == default:
        LOG.debug(f"Set default inicfg: {key} = {value!r}")


class TestRunnerTimeoutManager(tobiko.SharedFixture):
    timeout: tobiko.Seconds = None
    deadline: tobiko.Seconds = None

    def setup_fixture(self):
        tobiko_config = tobiko.tobiko_config()
        self.timeout = tobiko_config.testcase.test_runner_timeout
        if self.timeout is None:
            LOG.info('Test runner timeout is disabled')
        else:
            LOG.info('Test runner timeout is enabled: '
                     f'timeout is {self.timeout} seconds')
            self.deadline = tobiko.time() + self.timeout

    @classmethod
    def check_test_runner_timeout(cls):
        self = tobiko.setup_fixture(cls)
        if self.deadline is not None:
            time_left = self.deadline - tobiko.time()
            if time_left <= 0.:
                pytest.skip(
                    f"Test runner execution timed out after {self.timeout} "
                    f"seconds",
                    allow_module_level=True)
            else:
                LOG.debug('Test runner timeout is enabled: '
                          f'{time_left} seconds left')


def check_test_runner_timeout():
    TestRunnerTimeoutManager.check_test_runner_timeout()


def configure_timeout(config):
    tobiko_config = tobiko.tobiko_config()
    default = tobiko_config.testcase.timeout
    if default is not None and default > 0.:
        set_default_inicfg(config, 'timeout', default)


def pytest_html_results_table_header(cells):
    cells.insert(2, html.th("Description"))
    cells.insert(1, html.th("Time", class_="sortable time", col="time"))
    cells.pop()


def pytest_html_results_table_row(report, cells):
    cells.insert(2, html.td(getattr(report, 'description', '')))
    cells.insert(1, html.td(datetime.utcnow(), class_="col-time"))
    cells.pop()


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # pylint: disable=unused-argument
    outcome = yield
    report = outcome.get_result()
    report.description = getattr(item.function, '__doc__', '')


def pytest_html_report_title(report):
    report.title = f"Tobiko test results ({REPORT_NAME})"


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    # pylint: disable=unused-argument
    check_test_runner_timeout()
    yield


@pytest.fixture(scope="session", autouse=True)
def cleanup_shelves():
    tobiko.initialize_shelves()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from keystoneclient.v2_0 import client as v2_client
from keystoneclient.v3 import client as v3_client
from oslo_log import log
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.shell import sh


LOG = log.getLogger(__name__)

CIENT_CLASSSES = v2_client.Client, v3_client.Client


@keystone.skip_unless_has_keystone_credentials()
class TobikoKeystoneCredentialsCommandTest(testtools.TestCase):

    def test_execute(self):
        with sh.local_process('tobiko-keystone-credentials') as process:
            actual = tobiko.load_yaml(process.stdout)
        process.check_exit_status()
        expected = keystone.default_keystone_credentials().to_dict()
        self.assertEqual(expected, actual)


@keystone.skip_unless_has_keystone_credentials()
class KeystoneClientAPITest(testtools.TestCase):

    def test_get_keystone_client(self):
        client = keystone.get_keystone_client()
        self.assertIsInstance(client, CIENT_CLASSSES)

    def test_list_services(self):
        services = keystone.list_services()
        self.assertTrue(services)

    def test_list_services_by_name(self):
        services = keystone.list_services(name='keystone')
        self.assertTrue(services)
        for s in services:
            self.assertEqual('keystone', s.name)

    def test_list_services_by_type(self):
        services = keystone.list_services(type='identity')
        self.assertTrue(services)
        for s in services:
            self.assertEqual('identity', s.type)

    def test_find_service(self):
        service = keystone.find_service()
        self.assertTrue(service.id)

    def test_find_service_with_unique(self):
        self.assertRaises(tobiko.MultipleObjectsFound,
                          keystone.find_service,
                          unique=True)

    def test_find_service_not_found(self):
        self.assertRaises(tobiko.ObjectNotFound,
                          keystone.find_service,
                          name='never-never-land')

    def test_find_service_with_default(self):
        service = keystone.find_service(name='never-never-land',
                                        default=None)
        self.assertIsNone(service)

    def test_find_service_by_name(self):
        service = keystone.find_service(name='keystone')
        self.assertEqual('keystone', service.name)

    def test_find_service_by_type(self):
        service = keystone.find_service(type='identity')
        self.assertEqual('identity', service.type)

    def test_list_endpoints(self):
        service = keystone.find_service(name='keystone')
        endpoints = keystone.list_endpoints()
        self.assertIn(service.id, [e.service_id for e in endpoints])

    def test_list_endpoints_by_service(self):
        service = keystone.find_service(name='keystone')
        endpoints = keystone.list_endpoints(service=service)
        self.assertTrue(endpoints)
        self.assertEqual([service.id] * len(endpoints),
                         [e.service_id for e in endpoints])

    def test_list_endpoints_by_service_id(self):
        service = keystone.find_service(name='keystone')
        endpoints = keystone.list_endpoints(service_id=service.id)
        self.assertTrue(endpoints)
        for e in endpoints:
            self.assertEqual(service.id, e.service_id)

    def test_list_endpoints_by_interface(self):
        endpoints = keystone.list_endpoints(interface='public')
        self.assertTrue(endpoints)
        for e in endpoints:
            self.assertEqual('public', e.interface)

    def test_list_endpoints_by_url(self):
        url = keystone.list_endpoints()[-1].url
        endpoints = keystone.list_endpoints(url=url)
        self.assertTrue(endpoints)
        for e in endpoints:
            self.assertEqual(url, e.url)

    def test_find_endpoint(self):
        endpoint = keystone.find_endpoint()
        self.assertTrue(endpoint.id)

    def test_find_endpoint_with_unique(self):
        self.assertRaises(tobiko.MultipleObjectsFound,
                          keystone.find_endpoint,
                          unique=True)

    def test_find_endpoint_not_found(self):
        self.assertRaises(tobiko.ObjectNotFound,
                          keystone.find_endpoint,
                          service='never-never-land')

    def test_find_endpoint_with_default(self):
        service = keystone.find_endpoint(service='never-never-land',
                                         default=None)
        self.assertIsNone(service)

    def test_find_endpoint_by_service(self):
        service = keystone.find_service(name='keystone')
        endpoint = keystone.find_endpoint(service=service)
        self.assertEqual(endpoint.service_id, service.id)

    def test_find_endpoint_by_service_id(self):
        service = keystone.find_service(name='keystone')
        endpoint = keystone.find_endpoint(service_id=service.id)
        self.assertEqual(endpoint.service_id, service.id)

    def test_find_endpoint_by_url(self):
        url = keystone.list_endpoints()[-1].url
        endpoint = keystone.find_endpoint(url=url)
        self.assertEqual(url, endpoint.url)

    def test_find_service_endpoint(self):
        service = keystone.find_service(name='keystone')
        endpoint = keystone.find_service_endpoint(name='keystone')
        self.assertEqual(service.id, endpoint.service_id)
        self.assertEqual('public', endpoint.interface)
        self.assertTrue(endpoint.enabled)

    @keystone.skip_if_missing_service(name='octavia')
    def test_find_octavia_service_endpoint(self):
        service = keystone.find_service(name='octavia')
        endpoint = keystone.find_service_endpoint(name='octavia')
        self.assertEqual(service.id, endpoint.service_id)
        self.assertEqual('public', endpoint.interface)
        self.assertTrue(endpoint.enabled)

    def test_get_session_endpoint(self):
        endpoint = keystone.get_keystone_endpoint(
            service_type='identity')
        self.assertIsInstance(endpoint, str)

    def test_get_session_token(self):
        token = keystone.get_keystone_token()
        self.assertIsInstance(token, str)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc
import asyncio
import inspect
import logging
import typing
import uuid

from oslo_log import log

import tobiko
from tobiko.actors import _proxy
from tobiko.actors import _request


A = typing.TypeVar('A', bound='ActorBase')


class ActorRef(_proxy.CallProxyBase, _proxy.Generic[A]):

    _actor_class: typing.Type[A]

    def __class_getitem__(cls, item):
        # pylint: disable=too-many-function-args
        if isinstance(item, type):
            if issubclass(item, ActorBase):
                actor_class: typing.Type[ActorBase] = item
                base_name = cls.__name__.split('[', 1)[0]
                ref_class = _proxy.create_call_proxy_class(
                    protocols=(actor_class,),
                    class_name=f"{base_name}[{actor_class.__name__}]",
                    bases=(cls,),
                    namespace=dict(_actor_class=item),
                    predicate=is_actor_method)
                ref_class.__module__ = cls.__module__
                return ref_class
            else:
                raise TypeError(f'{item} is not subclass of {ActorBase}')

        ref_class = cls
        getitem = getattr(super(), '__class_getitem__')
        if callable(getitem):
            if inspect.ismethod(getitem):
                ref_class = getitem(item)
            else:
                ref_class = getitem(cls, item)
        return ref_class

    def __init__(self, actor: A):
        super().__init__()
        self._actor = tobiko.check_valid_type(actor, self._actor_class)

    def _handle_call(self, method: typing.Callable, *args, **kwargs) \
            -> asyncio.Future:
        arguments = inspect.signature(method).bind(
            None, *args, **kwargs).arguments
        arguments.pop('self', None)
        return self._actor.send_request(method.__name__, **arguments)

    def __repr__(self):
        return f'{type(self).__name__}({self._actor})'


def is_actor_method(obj):
    return getattr(obj, '__tobiko_actor_method__', False)


def actor_method(obj):
    if not callable(obj):
        raise TypeError(f"Actor method {obj} is not callable")

    if not inspect.iscoroutinefunction(obj):
        raise TypeError(f"Actor method {obj} is not async")

    if not _proxy.is_public_function(obj):
        raise TypeError(f"Actor method name {obj} can't start with '_'")

    name = getattr(obj, '__name__', None)
    if name is None or hasattr(ActorRef, name) or hasattr(Actor, name):
        raise TypeError(f"Invalid method name: '{name}'")

    obj.__tobiko_actor_method__ = True
    return obj


class _DummyActorProtocol(abc.ABC):
    pass


class ActorBase(tobiko.SharedFixture):
    max_queue_size: int = 0

    # Class methods ----------------------------------------------------------

    def __init_subclass__(cls,
                          *args,
                          **kwargs):
        super().__init_subclass__(*args, **kwargs)
        cls._actor_ref_class = ActorRef[cls]

    def __init__(self,
                 actor_id: str = None,
                 loop: asyncio.AbstractEventLoop = None,
                 log: logging.LoggerAdapter = None,
                 requests: _request.ActorRequestQueue = None):
        # pylint: disable=redefined-outer-name
        super().__init__()
        self.actor_id = actor_id
        if log is None:
            log = self._init_log()
        self.log = log
        if loop is None:
            loop = self._init_loop()
        self.loop = loop
        if requests is None:
            requests = self._init_requests()
        self.requests = requests
        self.setup_actor_future = self.loop.create_future()
        self.cleanup_actor_future = self.loop.create_future()
        self.cleanup_actor_future.set_result(None)

    def setup_fixture(self):
        if self.actor_id is None:
            self.actor_id = self._setup_actor_id()
        if self.setup_actor_future.done():
            self.setup_actor_future.cancel()
            self.setup_actor_future = self.loop.create_future()
        if self.cleanup_actor_future.done():
            self.cleanup_actor_future.cancel()
            self.cleanup_actor_future = self.loop.create_future()

    @property
    def ref(self) -> 'ActorRef':
        return self._actor_ref_class(actor=self)

    @property
    def actor_name(self) -> str:
        return tobiko.get_fixture_name(self)

    def send_request(self, method: str, **arguments) -> asyncio.Future:
        if self.actor_id is None:
            raise ValueError("Actor not set up yet")
        return self.requests.send_request(actor_id=self.actor_id,
                                          method=method,
                                          arguments=arguments)

    # Private instance methods -----------------------------------------------

    def _init_log(self):
        return log.getLogger(self.actor_name)

    @staticmethod
    def _init_loop() -> asyncio.AbstractEventLoop:
        return asyncio.get_event_loop()

    def _init_requests(self) -> _request.ActorRequestQueue:
        return _request.create_request_queue(max_size=self.max_queue_size,
                                             loop=self.loop)

    @staticmethod
    def _setup_actor_id() -> str:
        return str(uuid.uuid4())


class Actor(ActorBase):

    _stop_actor = False
    _run_actor_task: asyncio.Task

    # Class methods ----------------------------------------------------------

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        cls._actor_methods = dict(inspect.getmembers(cls, is_actor_method))

    @classmethod
    def get_fixture_manager(cls) -> tobiko.FixtureManager:
        from tobiko.actors import _manager
        return _manager.actor_manager()

    # Public methods ---------------------------------------------------------

    def setup_fixture(self):
        super().setup_fixture()
        self._run_actor_task = self.loop.create_task(
            self._run_actor())

    def cleanup_fixture(self):
        super().cleanup_fixture()
        self._stop_actor = True
        if hasattr(self, '_run_actor_task'):
            if not self._run_actor_task.done():
                # must weak up the actor with a message
                self.ref.ping_actor('cleanup')

    async def setup_actor(self):
        pass

    async def cleanup_actor(self):
        pass

    async def ping_actor(self, data: typing.Any = None) -> typing.Any:
        return data
    ping_actor.__tobiko_actor_method__ = True  # type: ignore[attr-defined]

    # Private methods --------------------------------------------------------

    async def _run_actor(self):
        await self._setup_actor()
        self._stop_actor = False
        while not self._stop_actor:
            request = await self.requests.receive_request()
            await self._receive_request(request)
        with tobiko.exc_info(reraise=True):
            await self._cleanup_actor()

    async def _setup_actor(self):
        try:
            self.log.debug(f'Actor setup started {self.actor_name}')
            await self.setup_actor()
        except Exception as ex:
            self.log.exception(
                f'Failed Setting up actor: {self.actor_name} '
                f'({self.actor_id})')
            self.setup_actor_future.set_exception(ex)
        else:
            self.setup_actor_future.set_result(None)
            self.log.debug(f'Actor setup succeeded: {self.actor_name} '
                           f'({self.actor_id}).')

    async def _cleanup_actor(self):
        try:
            self.log.debug(f'Actor cleanup started: {self.actor_name} '
                           f'({self.actor_id}).')
            await self.cleanup_actor()
        except Exception as ex:
            self.cleanup_actor_future.set_exception(ex)
            self.log.exception(
                f'Actor cleanup failed: {self.actor_name} '
                f'({self.actor_id}).')
        else:
            self.cleanup_actor_future.set_result(None)
            self.log.debug(f'Actor cleanup succeeded: {self.actor_name} '
                           f'({self.actor_id}).')
        finally:
            with tobiko.exc_info(reraise=True):
                await self.requests.cancel_requests(actor_id=self.actor_id)

    async def _receive_request(self, request: _request.ActorRequest):
        tobiko.check_valid_type(request, _request.ActorRequest)
        if request.actor_id != self.actor_id:
            raise ValueError(f"Invalid request actor_id: {request.actor_id}")
        method = self._get_actor_method(request.method)
        try:
            result = await method(self, **request.arguments)
        except Exception as ex:
            request.future.set_exception(ex)
        else:
            request.future.set_result(result)

    def _get_actor_method(self, name: str) -> typing.Callable:
        method = self._actor_methods.get(name)
        if method is None:
            raise ValueError(f"Invalid request method name: {name}")
        return method



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools

from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.shiftstack import _keystone
from tobiko import tripleo

LOG = log.getLogger(__name__)


def check_shiftstack():
    try:
        tripleo.check_overcloud()
    except tripleo.OvercloudNotFound as ex:
        raise ShiftstackNotFound(
            reason=f'Overcloud not found ({ex})') from ex

    try:
        _keystone.shiftstack_keystone_session()
    except keystone.NoSuchKeystoneCredentials as ex:
        raise ShiftstackNotFound(
            reason=f'Keystone credentials not found ({ex})') from ex


class ShiftstackNotFound(tobiko.ObjectNotFound):
    message = 'shiftstack not found: {reason}'


@functools.lru_cache()
def has_shiftstack() -> bool:
    try:
        check_shiftstack()
    except ShiftstackNotFound as ex:
        LOG.debug(f'Shiftstack not found: {ex}')
        return False
    else:
        LOG.debug('Shiftstack found')
        return True


def skip_unless_has_shiftstack():
    return tobiko.skip_unless("Shifstack credentials not found",
                              predicate=has_shiftstack)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.tests.unit import openstack


V2_PARAMS = {
    'api_version': 2,
    'project_name': 'demo',
    'username': 'demo',
    'password': 'super-secret',
    'auth_url': 'http://10.0.0.1:5678/v2.0'}

V2_ENVIRON = {
    'OS_PROJECT_NAME': 'demo',
    'OS_USERNAME': 'demo',
    'OS_PASSWORD': 'super-secret',
    'OS_AUTH_URL': 'http://10.0.0.1:5678/v2.0'}

V2_ENVIRON_WITH_TENANT_NAME = {
    'OS_TENANT_NAME': 'demo',
    'OS_USERNAME': 'demo',
    'OS_PASSWORD': 'super-secret',
    'OS_AUTH_URL': 'http://10.0.0.1:5678/v2.0'}

V2_ENVIRON_WITH_VERSION = dict(V2_ENVIRON, OS_IDENTITY_API_VERSION='2')


V3_PARAMS = {
    'api_version': 3,
    'project_name': 'demo',
    'username': 'demo',
    'password': 'super-secret',
    'auth_url': 'http://10.0.0.1:5678/v3',
    'user_domain_name': 'Default',
    'project_domain_name': 'Default'}

V3_ENVIRON = {
    'OS_PROJECT_NAME': 'demo',
    'OS_USERNAME': 'demo',
    'OS_PASSWORD': 'super-secret',
    'OS_AUTH_URL': 'http://10.0.0.1:5678/v3',
    'OS_USER_DOMAIN_NAME': 'Default',
    'OS_PROJECT_DOMAIN_NAME': 'Default'}

V3_ENVIRON_WITH_VERSION = dict(V3_ENVIRON, OS_IDENTITY_API_VERSION='3')


def make_credentials(params, **kwargs):
    if kwargs:
        params = dict(params, **kwargs)
    return keystone.keystone_credentials(**params)


class KeystoneCredentialsTest(openstack.OpenstackTest):

    def test_validate_from_params_v2(self):
        credentials = make_credentials(V2_PARAMS)
        credentials.validate()
        self.assertEqual(V2_PARAMS, credentials.to_dict())
        self.assertEqual(
            "keystone_credentials("
            "api_version=2, "
            "auth_url='http://10.0.0.1:5678/v2.0', "
            "password='***', "
            "project_name='demo', "
            "username='demo')",
            repr(credentials))

    def test_validate_from_params_v3(self):
        credentials = make_credentials(V3_PARAMS)
        credentials.validate()
        self.assertEqual(V3_PARAMS, credentials.to_dict())
        self.assertEqual(
            "keystone_credentials("
            "api_version=3, "
            "auth_url='http://10.0.0.1:5678/v3', "
            "password='***', "
            "project_domain_name='Default', "
            "project_name='demo', "
            "user_domain_name='Default', "
            "username='demo')",
            repr(credentials))

    def test_validate_without_auth_url(self):
        credentials = make_credentials(V2_PARAMS, auth_url=None)
        self.assertRaises(keystone.InvalidKeystoneCredentials,
                          credentials.validate)

    def test_validate_without_username(self):
        credentials = make_credentials(V2_PARAMS, username=None)
        self.assertRaises(keystone.InvalidKeystoneCredentials,
                          credentials.validate)

    def test_validate_without_project_name(self):
        credentials = make_credentials(V2_PARAMS, project_name=None)
        self.assertRaises(keystone.InvalidKeystoneCredentials,
                          credentials.validate)

    def test_validate_without_password(self):
        credentials = make_credentials(V2_PARAMS, password=None)
        self.assertRaises(keystone.InvalidKeystoneCredentials,
                          credentials.validate)


class EnvironKeystoneCredentialsFixtureTest(openstack.OpenstackTest):

    def test_init(self):
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        self.assertIsNone(fixture.credentials)

    def test_setup_with_no_credentials(self):
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        self.assertRaises(keystone.NoSuchKeystoneCredentials,
                          tobiko.setup_fixture,
                          fixture)

    def test_setup_v2(self):
        self.patch(os, 'environ', V2_ENVIRON)
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_v2_with_tenant_name(self):
        self.patch(os, 'environ', V2_ENVIRON_WITH_TENANT_NAME)
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_v2_with_api_version(self):
        self.patch(os, 'environ', V2_ENVIRON_WITH_VERSION)
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_v3(self):
        self.patch(os, 'environ', V3_ENVIRON)
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V3_PARAMS, fixture.credentials.to_dict())

    def test_setup_v3_without_api_version(self):
        self.patch(os, 'environ', V3_ENVIRON_WITH_VERSION)
        fixture = keystone.EnvironKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V3_PARAMS, fixture.credentials.to_dict())


class ConfigKeystoneCredentialsFixtureTest(openstack.OpenstackTest):

    def patch_config(self, params, **kwargs):
        credentials = make_credentials(params, **kwargs)
        return self.patch(config.CONF.tobiko, 'keystone', credentials)

    def test_init(self):
        fixture = keystone.ConfigKeystoneCredentialsFixture()
        self.assertIsNone(fixture.credentials)

    def test_setup_v2(self):
        self.patch_config(V2_PARAMS, api_version=None)
        fixture = keystone.ConfigKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_v2_with_api_version(self):
        self.patch_config(V2_PARAMS, api_version=2)
        fixture = keystone.ConfigKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_v3(self):
        self.patch_config(V3_PARAMS, api_version=None)
        fixture = keystone.ConfigKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V3_PARAMS, fixture.credentials.to_dict())

    def test_setup_v3_with_api_version(self):
        self.patch_config(V3_PARAMS, api_version=3)
        fixture = keystone.ConfigKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V3_PARAMS, fixture.credentials.to_dict())


class DelegateKeystoneCredentialsFixtureTest(openstack.OpenstackTest):

    def setUp(self):
        super().setUp()
        self.patch_config({})
        self.patch(os, 'environ', {})
        tobiko.remove_fixture(keystone.ConfigKeystoneCredentialsFixture)
        tobiko.remove_fixture(keystone.EnvironKeystoneCredentialsFixture)

    def patch_config(self, params, **kwargs):
        credentials = make_credentials(params, **kwargs)
        keystone_conf = tobiko.tobiko_config().keystone
        for name, value in credentials.to_dict().items():
            self.patch(keystone_conf, name, value)
        tobiko.tobiko_config().ssh.proxy_jump = None

    def test_init(self):
        fixture = keystone.DelegateKeystoneCredentialsFixture()
        self.assertIsNone(fixture.credentials)

    def test_setup_from_config(self):
        self.patch_config(V2_PARAMS)
        fixture = keystone.DelegateKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_from_environ(self):
        self.patch(os, 'environ', V2_ENVIRON)
        fixture = keystone.DelegateKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V2_PARAMS, fixture.credentials.to_dict())

    def test_setup_from_environ_and_config(self):
        self.patch(os, 'environ', V3_ENVIRON)
        self.patch_config(V2_PARAMS)
        fixture = keystone.DelegateKeystoneCredentialsFixture()
        fixture.setUp()
        fixture.credentials.validate()
        self.assertEqual(V3_PARAMS, fixture.credentials.to_dict())


class SkipUnlessHasKeystoneCredentialsTest(openstack.OpenstackTest):

    def patch_has_keystone_credentials(self, return_value: bool):
        # pylint: disable=protected-access
        from tobiko.openstack.keystone import _credentials
        return self.patch(_credentials,
                          'has_keystone_credentials',
                          return_value=return_value)

    def test_skip_method_unless_has_keystone_credentials(self):
        has_keystone_credentials = self.patch_has_keystone_credentials(True)

        call_args = []

        @keystone.skip_unless_has_keystone_credentials()
        def decorated_func(*args, **kwargs):
            call_args.append([args, kwargs])

        has_keystone_credentials.assert_not_called()

        decorated_func(1, 2, a=1, b=2)

        # pylint: disable=no-member
        has_keystone_credentials.assert_called_once()
        self.assertEqual(call_args, [[(1, 2), {'a': 1, 'b': 2}]])

    def test_skip_class_unless_has_keystone_credentials(self):
        has_keystone_credentials = self.patch_has_keystone_credentials(True)

        calls = []

        @keystone.skip_unless_has_keystone_credentials()
        class SkipTest(testtools.TestCase):

            def test_skip(self):
                calls.append(True)

        has_keystone_credentials.assert_not_called()

        tobiko.run_test(SkipTest('test_skip'))

        has_keystone_credentials.assert_called_once()
        self.assertEqual(calls, [True])

    def test_skip_method_unless_has_keystone_credentials_without_creds(self):
        has_keystone_credentials = self.patch_has_keystone_credentials(False)

        @keystone.skip_unless_has_keystone_credentials()
        def decorated_func():
            self.fail('Not skipped')

        has_keystone_credentials.assert_not_called()

        self.assertRaises(self.skipException, decorated_func)
        has_keystone_credentials.assert_called_once()

    def test_skip_class_unless_has_keystone_credentials_without_creds(self):
        has_keystone_credentials = self.patch_has_keystone_credentials(False)

        @keystone.skip_unless_has_keystone_credentials()
        class SkipTest(testtools.TestCase):

            def test_skip(self):
                super(SkipTest, self).setUp()
                self.fail('Not skipped')

        case = SkipTest('test_skip')
        has_keystone_credentials.assert_not_called()

        result = tobiko.run_test(case=case)
        tobiko.assert_test_case_was_skipped(
            case,
            result=result,
            skip_reason='Missing Keystone credentials')
        has_keystone_credentials.assert_called_once()



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
import collections

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


@keystone.skip_if_missing_service(name='octavia')
class OctaviaServicesFaultTest(testtools.TestCase):
    """Octavia services fault test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balancer while
    stopping some of the Octavia's services which run under podman container
    runtime environment.

    Each service will be running as a single instance.
    E.g. only one WORKER_SERVICE will run in all controllers, the same for
    API_SERVICE etc.

    Then we test that traffic which is being sent from the client to the LB
    is received as expected.
    """
    loadbalancer_stack = tobiko.required_fixture(
        stacks.AmphoraIPv4LoadBalancerStack)

    listener_stack = tobiko.required_fixture(
        stacks.HttpRoundRobinAmphoraIpv4Listener)

    # ssh clients of the participating TripleO nodes
    ssh_clients: typing.List[ssh.SSHClientFixture] = list()

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaServicesFaultTest, self).setUp()

        # Skipping the test if there are not enough instances of the services
        # (if there is only 1 controller and 1 networker, we cannot stop any
        # Octavia service)
        critical_nodes_number = len(
            topology.list_openstack_nodes(group='controller'))
        try:
            networker_nodes = topology.list_openstack_nodes(group='networker')
            critical_nodes_number += len(networker_nodes)
        except topology.NoSuchOpenStackTopologyNodeGroup:
            pass  # Current Octavia architecture doesn't support networker node

        if critical_nodes_number < 2:
            skip_reason = "The number of controllers and networker should be" \
                          " more than 1 for this test, otherwise each " \
                          "service is necessary."
            self.skipTest(skip_reason)

        # Wait for Octavia objects to be active
        LOG.info('Waiting for member '
                 f'{self.listener_stack.server_stack.stack_name} and '
                 f'for member '
                 f'{self.listener_stack.other_server_stack.stack_name} '
                 f'to be created...')
        self.listener_stack.wait_for_active_members()

        self.loadbalancer_stack.wait_for_octavia_service()

        self.listener_stack.wait_for_members_to_be_reachable()

        # For 5 minutes we ignore specific exceptions as we know
        # that Octavia resources are being provisioned
        # Sending initial traffic before we stop octavia services
        for attempt in tobiko.retry(timeout=300.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.listener_stack.pool_id,
                    ip_address=self.loadbalancer_stack.floating_ip_address,
                    lb_algorithm=self.listener_stack.lb_algorithm,
                    protocol=self.listener_stack.lb_protocol,
                    port=self.listener_stack.lb_port)
                break
            except (octavia.RoundRobinException,
                    octavia.TrafficTimeoutError,
                    sh.ShellCommandFailed):
                LOG.exception(f"Traffic didn't reach all members after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise

    def test_services_fault(self):
        # We get the services we want to stop on all nodes, in a way we leave
        # only one instance of every service on all nodes accumulatively
        services_to_stop = self._get_services_to_stop()

        try:
            self._stop_octavia_main_services(services_to_stop)

        finally:
            self._start_octavia_main_services(services_to_stop)

    def _get_services_to_stop(self) -> dict:
        """Return the running octavia services on controller & networker nodes.

        This method returns a dictionary of the services we are going to stop
        as keys, and the nodes we are going to stop the services on as values.

        For example:
        {
            'tripleo_octavia_worker.service': [ssh_client_of_controller-0,
                                               ssh_client_of_controller-1],
            'tripleo_octavia_api.service': [ssh_client_of_controller-1,
                                            ssh_client_of_controller-2]
             and so on....
        }

        We are leaving exactly one running instance of each service on all
        nodes together.
        The Octavia service that will remain running will be chosen randomly.
        """

        # Gather all controller ssh clients
        for controller in topology.list_openstack_nodes(group='controller'):
            self.ssh_clients.append(controller.ssh_client)

        # Gather all networker ssh clients if reachable
        # (for Composable and Upgrade jobs)
        try:
            for networker in topology.list_openstack_nodes(
                    group='networker'):
                self.ssh_clients.append(networker.ssh_client)
        except topology.NoSuchOpenStackTopologyNodeGroup:
            pass

        # Creating initial mapping of Octavia active units (services) which are
        # currently running on the nodes we gathered above
        services_on_nodes = collections.defaultdict(list)

        # Gathering all Octavia active units (services) which are currently
        # running on the nodes we gathered above
        for ssh_client in self.ssh_clients:
            for service in sh.list_systemd_units(ssh_client=ssh_client):
                if service.unit in octavia.OCTAVIA_SERVICES:
                    services_on_nodes[service.unit].append(ssh_client)

        # Example of the curret services_on_nodes:
        # {
        #     'tripleo_octavia_worker.service': [ssh_client_of_controller-0,
        #                                        ssh_client_of_controller-1],
        #     'tripleo_octavia_api.service': [ssh_client_of_controller-1,
        #                                     ssh_client_of_controller-2]
        #      and so on....
        # }

        LOG.debug(f'Full services_on_nodes dictionary:\n {services_on_nodes}')

        # Pop one random node's ssh client from each service.
        # We do that so we could leave exactly one single instance of each
        # service running on all nodes
        import random
        for service in services_on_nodes:
            nodes_length = len(services_on_nodes[service])
            services_on_nodes[service].pop(random.randint(0, nodes_length - 1))

        return services_on_nodes

    def _stop_octavia_main_services(self, services_to_stop: dict):

        """Stops the provided octavia services.

        This method stops the provided octavia services.

        It then sends traffic to validate the Octavia's functionality
        """

        LOG.debug(f'Services we are going to stop:\n {services_to_stop}')

        for service, ssh_clients in services_to_stop.items():
            for ssh_client in ssh_clients:
                sh.stop_systemd_units(service,
                                      ssh_client=ssh_client,
                                      sudo=True)
                LOG.debug(f'We stopped {service} on {ssh_client.host}')

        self.loadbalancer_stack.wait_for_octavia_service()

        # For 30 seconds we ignore the OctaviaClientException as we know
        # that Octavia services are being stopped and restarted
        for attempt in tobiko.retry(timeout=30.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.listener_stack.pool_id,
                    ip_address=self.loadbalancer_stack.floating_ip_address,
                    lb_algorithm=self.listener_stack.lb_algorithm,
                    protocol=self.listener_stack.lb_protocol,
                    port=self.listener_stack.lb_port)
                break
            except octavia.OctaviaClientException:
                LOG.exception(f"Octavia service was unavailable after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise

    def _start_octavia_main_services(self, services_to_stop: dict):

        """Start the octavia services.

        This method starts the provided octavia services on the nodes we
        gathered before.

        It then sends traffic to validate the Octavia's functionality
        """

        for service, ssh_clients in services_to_stop.items():
            for ssh_client in ssh_clients:
                sh.start_systemd_units(service,
                                       ssh_client=ssh_client,
                                       sudo=True)

                LOG.debug(f'We started {service} on {ssh_client.host}')

        octavia.check_members_balanced(
            pool_id=self.listener_stack.pool_id,
            ip_address=self.loadbalancer_stack.floating_ip_address,
            lb_algorithm=self.listener_stack.lb_algorithm,
            protocol=self.listener_stack.lb_protocol,
            port=self.listener_stack.lb_port)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.curl import _execute
from tobiko.shell.curl import _process


execute_curl = _execute.execute_curl

CurlHeader = _process.CurlHeader
CurlProcessFixture = _process.CurlProcessFixture
assert_downloaded_file = _process.assert_downloaded_file
download_file = _process.download_file
default_download_dir = _process.default_download_dir
get_url_header = _process.get_url_header



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import dbm
import os
import shelve

from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)
TEST_RUN_SHELF = 'test_run'


def get_shelves_dir():
    # ensure the directory exists
    from tobiko import config
    shelves_dir = os.path.expanduser(config.CONF.tobiko.common.shelves_dir)
    return shelves_dir


def get_shelf_path(shelf):
    return os.path.join(get_shelves_dir(), shelf)


def addme_to_shared_resource(shelf, resource):
    shelf_path = get_shelf_path(shelf)
    # this is needed for unit tests
    resource = str(resource)
    testcase_id = tobiko.get_test_case().id()
    for attempt in tobiko.retry(timeout=10.0,
                                interval=0.5):
        try:
            with shelve.open(shelf_path) as db:
                if db.get(resource) is None:
                    db[resource] = set()
                # the add and remove methods do not work directly on the db
                auxset = db[resource]
                auxset.add(testcase_id)
                db[resource] = auxset
                return db[resource]
        except dbm.error:
            LOG.exception(f"Error accessing shelf {shelf}")
            if attempt.is_last:
                raise


def removeme_from_shared_resource(shelf, resource):
    shelf_path = get_shelf_path(shelf)
    # this is needed for unit tests
    resource = str(resource)
    testcase_id = tobiko.get_test_case().id()
    for attempt in tobiko.retry(timeout=10.0,
                                interval=0.5):
        try:
            with shelve.open(shelf_path) as db:
                # the add and remove methods do not work directly on the db
                db[resource] = db.get(resource) or set()
                if testcase_id in db[resource]:
                    auxset = db[resource]
                    auxset.remove(testcase_id)
                    db[resource] = auxset
                return db[resource]
        except dbm.error:
            LOG.exception(f"Error accessing shelf {shelf}")
            if attempt.is_last:
                raise


def remove_test_from_shelf_resources(testcase_id, shelf):
    shelf_path = get_shelf_path(shelf)
    for attempt in tobiko.retry(timeout=10.0,
                                interval=0.5):
        try:
            with shelve.open(shelf_path) as db:
                if not db:
                    return
                for resource in db.keys():
                    if testcase_id in db[resource]:
                        auxset = db[resource]
                        auxset.remove(testcase_id)
                        db[resource] = auxset
                return db
        except dbm.error as err:
            LOG.exception(f"Error accessing shelf {shelf}")
            if "db type could not be determined" in str(err):
                # remove the filename extension, which depends on the specific
                # DBM implementation
                shelf_path = '.'.join(shelf_path.split('.')[:-1])
            if attempt.is_last:
                raise


def remove_test_from_all_shared_resources(testcase_id):
    LOG.debug(f'Removing test {testcase_id} from all shelf resources')
    shelves_dir = get_shelves_dir()
    for filename in os.listdir(shelves_dir):
        if TEST_RUN_SHELF not in filename:
            remove_test_from_shelf_resources(testcase_id, filename)


def initialize_shelves():
    shelves_dir = get_shelves_dir()
    shelf_path = os.path.join(shelves_dir, TEST_RUN_SHELF)
    id_key = 'PYTEST_XDIST_TESTRUNUID'
    test_run_uid = os.environ.get(id_key)

    tobiko.makedirs(shelves_dir)

    # if no PYTEST_XDIST_TESTRUNUID ->
    #     pytest was executed with only one worker
    # if tobiko.initialize_shelves() == True ->
    #    this is the first pytest worker running cleanup_shelves
    # then, cleanup the shelves directory
    # else, another worker did it before
    for attempt in tobiko.retry(timeout=15.0,
                                interval=0.5):
        try:
            with shelve.open(shelf_path) as db:
                if test_run_uid is None:
                    LOG.debug("Only one pytest worker - Initializing shelves")
                elif test_run_uid == db.get(id_key):
                    LOG.debug("Another pytest worker already initialized "
                              "the shelves")
                    return
                else:
                    LOG.debug("Initializing shelves for the "
                              "test run uid %s", test_run_uid)
                    db[id_key] = test_run_uid
                for filename in os.listdir(shelves_dir):
                    if TEST_RUN_SHELF not in filename:
                        file_path = os.path.join(shelves_dir, filename)
                        os.unlink(file_path)
                return
        except dbm.error:
            LOG.exception(f"Error accessing shelf {TEST_RUN_SHELF}")
            if attempt.is_last:
                raise



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.ping import _assert
from tobiko.shell.ping import _exception
from tobiko.shell.ping import _interface
from tobiko.shell.ping import _parameters
from tobiko.shell.ping import _ping
from tobiko.shell.ping import _statistics


assert_reachable_hosts = _assert.assert_reachable_hosts
assert_unreachable_hosts = _assert.assert_unreachable_hosts

BadAddressPingError = _exception.BadAddressPingError
LocalPingError = _exception.LocalPingError
ConnectPingError = _exception.ConnectPingError
PingFailed = _exception.PingFailed
PingError = _exception.PingError
ReachableHostsException = _exception.ReachableHostsException
UnreachableHostsException = _exception.UnreachableHostsException
PingException = _exception.PingException
SendToPingError = _exception.SendToPingError
UnknowHostError = _exception.UnknowHostError

skip_if_missing_fragment_ping_option = (
    _interface.skip_if_missing_fragment_ping_option)
has_ping_fragment_option = _interface.has_fragment_ping_option

ping_parameters = _parameters.ping_parameters
get_ping_parameters = _parameters.get_ping_parameters
default_ping_parameters = _parameters.default_ping_parameters

list_reachable_hosts = _ping.list_reachable_hosts
list_unreachable_hosts = _ping.list_unreachable_hosts
ping = _ping.ping
ping_hosts = _ping.ping_hosts
ping_until_delivered = _ping.ping_until_delivered
ping_until_undelivered = _ping.ping_until_undelivered
ping_until_received = _ping.ping_until_received
ping_until_unreceived = _ping.ping_until_unreceived
wait_for_ping_hosts = _ping.wait_for_ping_hosts

TRANSMITTED = _ping.TRANSMITTED
UNDELIVERED = _ping.UNDELIVERED
DELIVERED = _ping.DELIVERED
RECEIVED = _ping.RECEIVED
UNRECEIVED = _ping.UNRECEIVED

PingStatistics = _statistics.PingStatistics
write_ping_to_file = _ping.write_ping_to_file
check_ping_statistics = _ping.check_ping_statistics
skip_check_ping_statistics = _ping.skip_check_ping_statistics



# Copyright (c) 2022 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import nova
from tobiko.openstack import stacks


LOG = log.getLogger(__name__)


class CrateDeleteServerStackFixture(stacks.CirrosServerStackFixture):
    """Nova server for testing server deletion and creation
    """


@pytest.mark.minimal
@config.skip_if_prevent_create()
class CrateDeleteServerStackTest(testtools.TestCase):

    stack = tobiko.required_fixture(CrateDeleteServerStackFixture,
                                    setup=False)

    @classmethod
    def tearDownClass(cls) -> None:
        tobiko.cleanup_fixture(cls.stack.fixture)

    def test_1_create_server(self):
        """Test Nova server creation
        """
        tobiko.cleanup_fixture(self.stack)
        self.ensure_server()
        self.stack.assert_is_reachable()

    def test_2_delete_server(self):
        server = self.ensure_server(status='ACTIVE')
        self.stack.assert_is_reachable()

        nova.delete_server(server.id)
        for _ in tobiko.retry(timeout=60., interval=3.):
            try:
                server = nova.get_server(server_id=server.id)
            except nova.ServerNotFoundError:
                LOG.debug(f"Server '{server.id}' deleted")
                break
            else:
                LOG.debug(f"Waiting for server deletion:\n"
                          f" - server.id='{server.id}'"
                          f" - server.status='{server.status}'")
        self.stack.assert_is_unreachable()

    def ensure_server(self, status='ACTIVE'):
        try:
            server_id: str = self.stack.server_id
            nova.get_server(server_id=server_id)
        except heat.HeatStackNotFound:
            tobiko.setup_fixture(self.stack)
        except nova.ServerNotFoundError:
            tobiko.reset_fixture(self.stack)
        return self.stack.ensure_server_status(status=status)



#!/usr/bin/env python3
# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import psutil
import resource
import signal
import sys
import subprocess


TOP_DIR = os.path.dirname(os.path.dirname(__file__))
if TOP_DIR not in sys.path:
    sys.path.insert(0, TOP_DIR)

from tools import common  # noqa

LOG = common.get_logger(__name__)

# Root tests dir
TEST_PATH = common.normalize_path(
    os.environ.get('TOBIKO_TEST_PATH') or
    os.environ.get('OS_TEST_PATH') or
    os.path.join(TOP_DIR, 'tobiko', 'tests', 'unit'))

# Output dirs
REPORT_DIR = common.normalize_path(
    os.environ.get('TOBIKO_REPORT_DIR') or
    os.environ.get('TOX_REPORT_DIR') or
    os.getcwd())

REPORT_NAME = (
    os.environ.get('TOBIKO_REPORT_NAME') or
    os.environ.get('TOX_REPORT_NAME') or
    'tobiko_results')

REPORT_PREFIX = os.path.join(REPORT_DIR, REPORT_NAME)

REPORT_LOG = (
    os.environ.get('TOBIKO_REPORT_LOG') or
    os.environ.get('TOX_REPORT_LOG') or
    REPORT_PREFIX + '.log')

REPORT_HTML = (
    os.environ.get('TOBIKO_REPORT_HTML') or
    os.environ.get('TOX_REPORT_HTML') or
    REPORT_PREFIX + '.html')

REPORT_XML = (
    os.environ.get('TOBIKO_REPORT_XML') or
    os.environ.get('TOX_REPORT_XML') or
    REPORT_PREFIX + '.xml')

NUM_PROCESSES = (
    os.environ.get('TOBIKO_NUM_PROCESSES') or
    os.environ.get('TOX_NUM_PROCESSES') or
    'auto')

RUN_TESTS_TIMEOUT = float(
    os.environ.get('TOBIKO_RUN_TESTS_TIMEOUT') or
    os.environ.get('TOX_RUN_TESTS_TIMEOUT') or
    0.)

RERUNS = int(
    os.environ.get('TOBIKO_RERUNS') or
    os.environ.get('TOX_RERUNS') or
    0)

RERUNS_DELAY = int(
    os.environ.get('TOBIKO_RERUNS_DELAY') or
    os.environ.get('TOX_RERUNS_DELAY') or
    5)

COVER = (
    os.environ.get('TOBIKO_COVER') or
    os.environ.get('TOX_COVER') or
    'false') in ['1', 'yes', 'true']


def main():
    common.setup_logging()
    try:
        succeeded = run_tests()
        if succeeded:
            LOG.info('SUCCEEDED')
            sys.exit(0)
        else:
            LOG.info('FAILED')
            sys.exit(1)

    except Exception:
        LOG.exception('ERROR')
        sys.exit(2)


def run_tests():
    setup_ulimits()
    setup_timeout()
    cleanup_report_dir()
    log_environ()

    succeeded = True
    try:
        run_test_cases()
    except (subprocess.CalledProcessError,
            subprocess.TimeoutExpired,
            ProcessLookupError) as ex:
        LOG.error(f"Error while running test cases.\n{ex}")
        succeeded = False

    return succeeded


def setup_ulimits():
    try:
        hard_nofile_limit = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        resource.setrlimit(
            resource.RLIMIT_NOFILE,
            (hard_nofile_limit, hard_nofile_limit))
    except ValueError:
        pass


def setup_timeout():

    if RUN_TESTS_TIMEOUT > 0.:

        def handle_timeout(_signum, _frame):
            LOG.error(
                f"run_tests.py timeout out after {RUN_TESTS_TIMEOUT} "
                "seconds")
            terminate_childs()
            raise subprocess.TimeoutExpired("run_tests.py",
                                            RUN_TESTS_TIMEOUT)

        signal.setitimer(signal.ITIMER_REAL, RUN_TESTS_TIMEOUT)
        signal.signal(signal.SIGALRM, handle_timeout)
        LOG.debug(f'Run tests timeout set as {RUN_TESTS_TIMEOUT} seconds')


def terminate_childs():
    current_process = psutil.Process()
    children = current_process.children(recursive=False)
    for child in children:
        LOG.debug(f"Interrupt child process execution (pid={child.pid})")
        os.kill(child.pid, signal.SIGINT)
    for child in children:
        LOG.debug("Wait for top-child process termination "
                  f"(pid={child.pid})...")
        os.waitpid(child.pid, 0)


def cleanup_report_dir():
    for report_file in [REPORT_LOG, REPORT_HTML, REPORT_XML]:
        if not common.make_dir(os.path.dirname(report_file)):
            common.remove_file(report_file)


def log_environ():
    common.execute('env | sort >> "{log_file}"',
                   log_file=REPORT_LOG,
                   capture_stdout=False)


def run_test_cases():
    xdist_options = ''
    if NUM_PROCESSES != '1':
        xdist_options = f"--numprocesses '{NUM_PROCESSES}' --dist loadscope"
    rerun_options = ''
    if RERUNS:
        rerun_options = f"--reruns '{RERUNS}' --reruns-delay '{RERUNS_DELAY}'"
    cover_options = ''
    if COVER:
        cover_options = f"--cov=tobiko"

    # Pass environment variables to pytest command
    environ = dict(os.environ, TOBIKO_REPORT_NAME=REPORT_NAME)
    common.execute("pytest -v "
                   f"{xdist_options} "
                   f"{rerun_options} "
                   f"{cover_options} "
                   f"--log-file={REPORT_LOG} "
                   f"--junitxml={REPORT_XML} "
                   f"--html={REPORT_HTML} --self-contained-html "
                   f"{common.get_posargs()}",
                   environ=environ,
                   capture_stdout=False)


if __name__ == '__main__':
    main()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg
from oslo_log import log

LOG = log.getLogger(__name__)

GROUP_NAME = 'ssh'
OPTIONS = [
    cfg.BoolOpt('debug',
                default=False,
                help=('Logout debugging messages of paramiko '
                      'library')),
    cfg.StrOpt('command',
               default='/usr/bin/ssh',
               help=('Default SSH client command')),
    cfg.StrOpt('port',
               default=None,
               help=('Default SSH port')),
    cfg.StrOpt('username',
               default=None,
               help=('Default SSH username')),
    cfg.ListOpt('config_files',
                default=['ssh_config', '.ssh/config'],
                help="Default user SSH configuration files"),
    cfg.ListOpt('key_file',
                default=['.ssh/id',
                         '~/.ssh/id_dsa',
                         '~/.ssh/id_rsa',
                         '~/.ssh/id_ecdsa',
                         '~/.ssh/id_ed25519'],
                help="Default SSH private key file(s) wildcard"),
    cfg.BoolOpt('allow_agent',
                default=False,
                help=("Set to False to disable connecting to the "
                      "SSH agent")),
    cfg.BoolOpt('compress',
                default=False,
                help="Set to True to turn on compression"),
    cfg.FloatOpt('timeout',
                 default=15.,
                 help="SSH connect timeout in seconds"),
    cfg.IntOpt('connection_attempts',
               default=120,
               help=("Maximum number of connection attempts to be tried "
                     "before timeout")),
    cfg.FloatOpt('connection_interval',
                 default=5.,
                 help=("Minimal seconds to wait between every "
                       "failed SSH connection attempt")),
    cfg.IntOpt('connection_timeout',
               default=200.,
               help=("Time before stopping retrying establishing an SSH "
                     "connection")),
    cfg.StrOpt('proxy_jump',
               default=None,
               help="Default SSH proxy server"),
    cfg.StrOpt('proxy_command',
               default=None,
               help="Default proxy command"),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]


def setup_tobiko_config(conf):
    paramiko_logger = log.getLogger('paramiko')
    if conf.ssh.debug:
        if not paramiko_logger.isEnabledFor(log.DEBUG):
            # Print paramiko debugging messages
            paramiko_logger.logger.setLevel(log.DEBUG)
    else:
        if paramiko_logger.isEnabledFor(log.ERROR):
            # Silence paramiko debugging messages
            paramiko_logger.logger.setLevel(log.FATAL)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg


GROUP_NAME = 'shiftstack'
OPTIONS = [
    # ShiftStack options
    cfg.StrOpt('local_clouds_file_path',
               default='.tobiko/shifstack/clouds.yaml',
               help="local clouds file path"),
    cfg.StrOpt('remote_clouds_file_path',
               default='~/clouds.yaml',
               help="remote clouds file path on undercloud-0 host"),
    cfg.StrOpt('cloud_name',
               default='shiftstack',
               help="Keystone credentials cloud name"),
    cfg.ListOpt('rcfile',
                default=['./shiftstackrc'],
                help="Path to the RC file used to populate OS_* environment "
                     "variables")
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class PortTest(testtools.TestCase):

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    def test_get_port_with_id(self):
        return self._test_get_port(port=self.stack.port_id)

    def test_get_port_with_details(self):
        return self._test_get_port(port=self.stack.port_details)

    def _test_get_port(self, port: neutron.PortIdType):
        observed = neutron.get_port(port)
        port_id = neutron.get_port_id(port)
        self.assertEqual(port_id, observed['id'])

    def test_create_port(self,
                         network: neutron.NetworkIdType = None) \
            -> neutron.PortType:
        port = neutron.create_port(name=self.id(), network=network)
        self.assertIsInstance(port['id'], str)
        self.assertNotEqual('', port['id'])
        self.assertEqual(self.id(), port['name'])
        self.assertEqual(port['id'], neutron.get_port(port)['id'])
        if network is not None:
            self.assertEqual(neutron.get_network_id(network),
                             port['network_id'])
        return port

    def test_create_port_with_network_id(self):
        network = neutron.create_network()
        self.test_create_port(network=network['id'])

    def test_create_port_with_network_details(self):
        network = neutron.create_network()
        self.test_create_port(network=network)

    def test_delete_port_with_id(self):
        port = self.test_create_port()
        neutron.delete_port(port['id'])
        self.assertRaises(neutron.NoSuchPort, neutron.get_port,
                          port['id'])

    def test_delete_port_with_details(self):
        port = self.test_create_port()
        neutron.delete_port(port)
        self.assertRaises(neutron.NoSuchPort, neutron.get_port,
                          port['id'])

    def test_delete_network_with_invalid_id(self):
        self.assertRaises(neutron.NoSuchPort,
                          neutron.delete_port, '<invalid-id>')

    def test_list_port_addresses(self, ip_version=None):
        port = neutron.find_port(device_id=self.stack.server_id)
        port_addresses = neutron.list_port_ip_addresses(
            port=port,
            ip_version=ip_version)
        server_addresses = nova.list_server_ip_addresses(
            server=self.stack.server_details,
            ip_version=ip_version,
            address_type='fixed')
        self.assertEqual(set(server_addresses), set(port_addresses))
        if ip_version:
            self.assertEqual(
                port_addresses.with_attributes(version=ip_version),
                port_addresses)

    def test_list_port_addresses_with_ipv4(self):
        self.test_list_port_addresses(ip_version=4)

    def test_list_port_addresses_with_ipv6(self):
        self.test_list_port_addresses(ip_version=6)

    def test_find_port_address_with_ip_version(self):
        port = neutron.find_port(device_id=self.stack.server_id)
        server_addresses = nova.list_server_ip_addresses(
            server=self.stack.server_details,
            address_type='fixed')
        for server_address in server_addresses:
            port_address = neutron.find_port_ip_address(
                port=port,
                ip_version=server_address.version,
                unique=True)
            self.assertEqual(server_address, port_address)

    def test_find_port_address_with_subnet_id(self):
        port = neutron.find_port(device_id=self.stack.server_id)
        for subnet in neutron.list_subnets(network_id=port['network_id']):
            port_address = neutron.find_port_ip_address(
                port=port, subnet=subnet, unique=True)
            cidr = netaddr.IPNetwork(subnet['cidr'])
            self.assertIn(port_address, cidr)



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import os
import time

import testtools
from oslo_log import log
import pandas

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko import tripleo
from tobiko.tripleo import containers


LOG = log.getLogger(__name__)


@tripleo.skip_if_missing_overcloud
class ContainersHealthTest(testtools.TestCase):
    # TODO(eolivare): refactor this class, because it replicates some code from
    # tobiko/tripleo/containers.py and its tests may be duplicating what
    # test_0vercloud_health_check already covers when it calls
    # containers.assert_all_tripleo_containers_running()

    @functools.lru_cache()
    def list_node_containers(self, ssh_client):
        """returns a list of containers and their run state"""
        return containers.get_container_runtime().\
            list_containers(ssh_client=ssh_client)

    def test_cinder_api(self):
        """check that all common tripleo containers are running"""
        containers.assert_containers_running('controller', ['cinder_api'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_rsync(self):
        containers.assert_containers_running('controller', ['swift_rsync'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_proxy(self):
        containers.assert_containers_running('controller', ['swift_proxy'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_object_updater(self):
        containers.assert_containers_running('controller',
                                             ['swift_object_updater'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_object_server(self):
        containers.assert_containers_running('controller',
                                             ['swift_object_server'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_object_replicator(self):
        containers.assert_containers_running('controller',
                                             ['swift_object_replicator'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_object_expirer(self):
        containers.assert_containers_running('controller',
                                             ['swift_object_expirer'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_object_auditor(self):
        containers.assert_containers_running('controller',
                                             ['swift_object_auditor'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_container_updater(self):
        containers.assert_containers_running('controller',
                                             ['swift_container_updater'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_container_server(self):
        containers.assert_containers_running('controller',
                                             ['swift_container_server'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_container_replicator(self):
        containers.assert_containers_running('controller',
                                             ['swift_container_replicator'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_container_auditor(self):
        containers.assert_containers_running('controller',
                                             ['swift_container_auditor'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_account_server(self):
        containers.assert_containers_running('controller',
                                             ['swift_account_server'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_account_replicator(self):
        containers.assert_containers_running('controller',
                                             ['swift_account_replicator'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_account_reaper(self):
        containers.assert_containers_running('controller',
                                             ['swift_account_reaper'])

    @tripleo.skip_if_ceph_rgw()
    def test_swift_account_auditor(self):
        containers.assert_containers_running('controller',
                                             ['swift_account_auditor'])

    def test_nova_vnc_proxy(self):
        containers.assert_containers_running('controller', ['nova_vnc_proxy'])

    def test_nova_scheduler(self):
        containers.assert_containers_running('controller', ['nova_scheduler'])

    def test_nova_metadata(self):
        containers.assert_containers_running('controller', ['nova_metadata'])

    def test_nova_conductor(self):
        containers.assert_containers_running('controller', ['nova_conductor'])

    def test_nova_api_cron(self):
        containers.assert_containers_running('controller', ['nova_api_cron'])

    def test_nova_api(self):
        containers.assert_containers_running('controller', ['nova_api'])

    def test_neutron_api(self):
        containers.assert_containers_running('controller', ['neutron_api'])

    def test_memcached(self):
        containers.assert_containers_running('controller', ['memcached'])

    def test_controller_logrotate_crond(self):
        containers.assert_containers_running('controller', ['logrotate_crond'])

    def test_keystone(self):
        containers.assert_containers_running('controller', ['keystone'])

    def test_controller_iscsid(self):
        containers.assert_containers_running('controller', ['iscsid'])

    def test_horizon(self):
        containers.assert_containers_running('controller', ['horizon'])

    def test_heat_engine(self):
        containers.assert_containers_running('controller', ['heat_engine'])

    def test_heat_api_cron(self):
        containers.assert_containers_running('controller', ['heat_api_cron'])

    def test_heat_api_cfn(self):
        containers.assert_containers_running('controller', ['heat_api_cfn'])

    def test_heat_api(self):
        containers.assert_containers_running('controller', ['heat_api'])

    def test_glance_api(self):
        containers.assert_containers_running('controller', ['glance_api'])

    def test_cinder_scheduler(self):
        containers.assert_containers_running('controller',
                                             ['cinder_scheduler'])

    def test_cinder_api_cron(self):
        containers.assert_containers_running('controller', ['cinder_api_cron'])

    def test_compute_iscsid(self):
        containers.assert_containers_running('compute', ['iscsid'])

    def test_compute_logrotate_crond(self):
        containers.assert_containers_running('compute', ['logrotate_crond'])

    def test_nova_compute(self):
        containers.assert_containers_running('compute', ['nova_compute'])

    def test_nova_libvirt(self):
        nova_libvirt = containers.get_libvirt_container_name()
        containers.assert_containers_running('compute', [nova_libvirt])

    def test_nova_migration_target(self):
        containers.assert_containers_running('compute',
                                             ['nova_migration_target'])

    def test_nova_virtlogd(self):
        containers.assert_containers_running('compute', ['nova_virtlogd'])

    def test_ovn_containers_running(self):
        containers.assert_ovn_containers_running()

    def test_equal_containers_state(self, expected_containers_list=None,
                                    timeout=120, interval=5,
                                    recreate_expected=False):
        """compare all overcloud container states with using two lists:
        one is current , the other some past list
        first time this method runs it creates a file holding overcloud
        containers' states: ~/expected_containers_list_df.csv'
        second time it creates a current containers states list and
        compares them, they must be identical"""
        # if we have a file or an explicit variable use that ,
        # otherwise  create and return
        if recreate_expected or (not (expected_containers_list or
                                      os.path.exists(
                                          containers.
                                          expected_containers_file))):
            containers.save_containers_state_to_file(containers.
                                                     list_containers())
            return
        elif expected_containers_list:
            expected_containers_list_df = pandas.DataFrame(
                containers.get_container_states_list(expected_containers_list),
                columns=['container_host', 'container_name',
                         'container_state'])
        elif os.path.exists(containers.expected_containers_file):
            expected_containers_list_df = pandas.read_csv(
                containers.expected_containers_file)
        failures = []
        error_info = 'Output explanation: left_only is the original state, ' \
                     'right_only is the new state'
        for _ in tobiko.retry(timeout=timeout):
            failures = []
            actual_containers_list_df = containers.list_containers_df()
            LOG.info('expected_containers_list_df: {} '.format(
                expected_containers_list_df.to_string(index=False)))
            LOG.info('actual_containers_list_df: {} '.format(
                actual_containers_list_df.to_string(index=False)))
            # execute a `dataframe` diff between the expected
            # and actual containers
            expected_containers_state_changed = \
                containers.dataframe_difference(expected_containers_list_df,
                                                actual_containers_list_df)
            # check for changed state containerstopology
            if not expected_containers_state_changed.empty:
                failures.append('expected containers changed state ! : '
                                '\n\n{}\n{}'.format(
                                 expected_containers_state_changed.
                                 to_string(index=False), error_info))
                LOG.info('container states mismatched:\n{}\n'.format(failures))
                time.sleep(interval)
                # clear cache to obtain new data
                containers.list_node_containers.cache_clear()
            else:
                LOG.info("assert_equal_containers_state :"
                         " OK, all containers are on the same state")
                return
        if failures:
            tobiko.fail('container states mismatched:\n{!s}', '\n'.join(
                failures))

    def config_validation(self, config_checkings):
        container_runtime_name = containers.get_container_runtime_name()
        for node in topology.list_openstack_nodes(
                group=config_checkings['node_group']):
            for param_check in config_checkings['param_validations']:
                obtained_param = sh.execute(
                    f"{container_runtime_name} exec -uroot "
                    f"{config_checkings['container_name']} crudini "
                    f"--get {config_checkings['config_file']} "
                    f"{param_check['section']} {param_check['param']}",
                    ssh_client=node.ssh_client, sudo=True).stdout.strip()
                self.assertTrue(param_check['expected_value'] in
                                obtained_param,
                                f"Expected {param_check['param']} value: "
                                f"{param_check['expected_value']}\n"
                                f"Obtained {param_check['param']} value: "
                                f"{obtained_param}")
        LOG.info("Configuration verified:\n"
                 f"node group: {config_checkings['node_group']}\n"
                 f"container: {config_checkings['container_name']}\n"
                 f"config file: {config_checkings['config_file']}")

    @neutron.skip_unless_is_ovn()
    def test_ovn_container_config(self):
        """check containers configuration in ovn
        """
        ovn_config_checkings = \
            {'node_group': 'controller',
             'container_name': 'neutron_api',
             'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
             'param_validations': [{'section': 'ml2',
                                    'param': 'mechanism_drivers',
                                    'expected_value': 'ovn'},
                                   {'section': 'ml2',
                                    'param': 'type_drivers',
                                    'expected_value': 'geneve'},
                                   {'section': 'ovn',
                                    'param': 'ovn_metadata_enabled',
                                    'expected_value': 'True'}]}
        self.config_validation(ovn_config_checkings)

    @neutron.skip_unless_is_ovs()
    def test_ovs_container_config(self):
        """check containers configuration in ovn
        """
        ovs_config_checkings = \
            {'node_group': 'controller',
             'container_name': 'neutron_api',
             'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
             'param_validations': [{'section': 'ml2',
                                    'param': 'mechanism_drivers',
                                    'expected_value': 'openvswitch'}]}
        self.config_validation(ovs_config_checkings)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import random

import netaddr
from netaddr.strategy import ipv4
from netaddr.strategy import ipv6

import tobiko
from tobiko.openstack.neutron import _client


def new_ipv4_cidr(seed=None):
    return tobiko.setup_fixture(IPv4CIDRGeneratorFixture).new_cidr(seed=seed)


def new_ipv6_cidr(seed=None):
    return tobiko.setup_fixture(IPv6CIDRGeneratorFixture).new_cidr(seed=seed)


class CIDRGeneratorFixture(tobiko.SharedFixture):

    cidr = None
    prefixlen = None
    client = None
    config = None
    cidr_generator = None

    def __init__(self, cidr=None, prefixlen=None, client=None):
        super(CIDRGeneratorFixture, self).__init__()
        if cidr:
            self.cidr = cidr
        if prefixlen:
            self.prefixlen = prefixlen
        if client:
            self.client = client

    def setup_fixture(self):
        self.setup_config()
        self.setup_client()

    def setup_config(self):
        from tobiko import config
        CONF = config.CONF
        self.config = CONF.tobiko.neutron

    def setup_client(self):
        self.client = _client.neutron_client(self.client)

    def new_cidr(self, seed):
        used_cidrs = set(list_subnet_cidrs(client=self.client))
        for cidr in random_subnets(cidr=self.cidr, prefixlen=self.prefixlen,
                                   seed=seed):
            if cidr not in used_cidrs:
                return cidr
        raise NoSuchCIDRLeft(cidr=self.cidr, prefixlen=self.prefixlen)


class IPv4CIDRGeneratorFixture(CIDRGeneratorFixture):

    @property
    def cidr(self):
        return netaddr.IPNetwork(self.config.ipv4_cidr)

    @property
    def prefixlen(self):
        return int(self.config.ipv4_prefixlen)


class IPv6CIDRGeneratorFixture(CIDRGeneratorFixture):

    @property
    def cidr(self):
        return netaddr.IPNetwork(self.config.ipv6_cidr)

    @property
    def prefixlen(self):
        return int(self.config.ipv6_prefixlen)


class NoSuchCIDRLeft(tobiko.TobikoException):
    message = ("No such subnet CIDR left "
               "(CIDR={cidr!s}, prefixlen={prefixlen!s})")


def random_subnets(cidr, prefixlen, seed=None):
    """
    A generator that divides up this IPNetwork's subnet into smaller
    subnets based on a specified CIDR prefix.

    :param prefixlen: a CIDR prefix indicating size of subnets to be
        returned.

    :return: an iterator containing random IPNetwork subnet objects.
    """

    version = cidr.version
    module = {4: ipv4, 6: ipv6}[version]
    width = module.width
    if not 0 <= cidr.prefixlen <= width:
        message = "CIDR prefix /{!r} invalid for IPv{!s}!".format(
            prefixlen, cidr.version)
        raise ValueError(message)

    if not cidr.prefixlen <= prefixlen:
        #   Don't return anything.
        raise StopIteration

    # Calculate number of subnets to be returned.
    max_subnets = 2 ** (width - cidr.prefixlen) // 2 ** (width - prefixlen)

    base_subnet = module.int_to_str(cidr.first)
    i = 0
    rand = random.Random(hash(seed) ^ os.getpid())
    while True:
        subnet = netaddr.IPNetwork('%s/%d' % (base_subnet, prefixlen), version)
        subnet.value += (subnet.size * rand.randrange(0, max_subnets))
        subnet.prefixlen = prefixlen
        i += 1
        yield subnet


def list_subnet_cidrs(client: _client.NeutronClientType = None,
                      **params) -> tobiko.Selection[netaddr.IPNetwork]:
    from tobiko.openstack.neutron import _subnet
    subnets = _subnet.list_subnets(client=client, **params)
    return tobiko.select(netaddr.IPNetwork(subnet['cidr'])
                         for subnet in subnets)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr

import tobiko
from tobiko.openstack.neutron import _client


NetworkType = typing.Dict[str, typing.Any]
NetworkIdType = typing.Union[str, NetworkType]


class NoSuchNetwork(tobiko.ObjectNotFound):
    message = "No such network found for {id!r}"


def get_network_id(network: NetworkIdType) -> str:
    if isinstance(network, str):
        return network
    else:
        return network['id']


def get_network(network: NetworkIdType,
                client: _client.NeutronClientType = None,
                **params) -> NetworkType:
    network_id = get_network_id(network)
    try:
        return _client.neutron_client(client).show_network(
            network_id, **params)['network']
    except _client.neutronclient.exceptions.NotFound as ex:
        raise NoSuchNetwork(id=network_id) from ex


def list_networks(client: _client.NeutronClientType = None,
                  **params) -> \
        tobiko.Selection[NetworkType]:
    networks = _client.neutron_client(client).list_networks(
        **params)['networks']
    return tobiko.select(networks)


def find_network(client: _client.NeutronClientType = None,
                 unique=False,
                 default: NetworkType = None,
                 **attributes) -> NetworkType:
    """Look for a network matching some property values"""
    networks = list_networks(client=client, **attributes)
    if default is None or networks:
        if unique:
            return networks.unique
        else:
            return networks.first
    else:
        return default


def create_network(client: _client.NeutronClientType = None,
                   add_cleanup=True,
                   **params) -> NetworkType:
    network = _client.neutron_client(client).create_network(
        body={'network': params})['network']
    if add_cleanup:
        tobiko.add_cleanup(cleanup_network, network=network, client=client)
    return network


def cleanup_network(network: NetworkIdType,
                    client: _client.NeutronClientType = None):
    try:
        delete_network(network=network, client=client)
    except NoSuchNetwork:
        pass


def delete_network(network: NetworkIdType,
                   client: _client.NeutronClientType = None):
    network_id = get_network_id(network)
    try:
        _client.neutron_client(client).delete_network(network=network_id)
    except _client.NotFound as ex:
        raise NoSuchNetwork(id=network_id) from ex


def list_network_nameservers(network_id: typing.Optional[str] = None,
                             ip_version: typing.Optional[int] = None) -> \
        tobiko.Selection[netaddr.IPAddress]:
    from tobiko.openstack.neutron import _subnet
    subnets = _subnet.list_subnets(network_id=network_id)
    nameservers = tobiko.Selection[netaddr.IPAddress](
        netaddr.IPAddress(nameserver)
        for subnets in subnets
        for nameserver in subnets['dns_nameservers'])
    if ip_version is not None:
        nameservers = nameservers.with_attributes(version=ip_version)
    return nameservers



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class FloatingIpTest(testtools.TestCase):

    server = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    @property
    def floating_network_id(self) -> str:
        return stacks.get_floating_network_id()

    @property
    def floating_network_details(self) -> neutron.NetworkType:
        return stacks.get_floating_network()

    def test_create_floating_ip(self,
                                network: neutron.NetworkIdType = None) -> \
            neutron.FloatingIpType:
        floating_ip = neutron.create_floating_ip(network=network)
        self.assertIsInstance(floating_ip, dict)
        if network is None:
            floating_network_id = self.floating_network_id
        else:
            floating_network_id = neutron.get_network_id(network)
        self.assertEqual(floating_network_id,
                         floating_ip['floating_network_id'])
        self.assertEqual(floating_ip, neutron.get_floating_ip(floating_ip))
        return floating_ip

    def test_create_floating_ip_with_network_id(self):
        self.test_create_floating_ip(network=self.floating_network_id)

    def test_create_floating_ip_with_network_details(self):
        self.test_create_floating_ip(network=self.floating_network_details)

    def test_delete_floating_ip_with_id(self):
        floating_ip = self.test_create_floating_ip()
        neutron.delete_floating_ip(floating_ip['id'])
        self.assertRaises(neutron.NoSuchFloatingIp,
                          neutron.get_floating_ip, floating_ip)

    def test_delete_floating_ip_with_details(self):
        floating_ip = self.test_create_floating_ip()
        neutron.delete_floating_ip(floating_ip)
        self.assertRaises(neutron.NoSuchFloatingIp,
                          neutron.get_floating_ip, floating_ip)

    def test_delete_floating_ip_with_invalid_id(self):
        self.assertRaises(neutron.NoSuchFloatingIp,
                          neutron.delete_floating_ip, '<invalid-id>')

    def test_list_floating_ips(self):
        port_id = self.server.port_id
        floating_ips = neutron.list_floating_ips()
        floating_ip = floating_ips.with_items(port_id=port_id).unique
        self.assertEqual(floating_ip['floating_ip_address'],
                         self.server.floating_ip_address)

    def test_list_floating_ip_with_port_id(self):
        port_id = self.server.port_id
        floating_ip = neutron.list_floating_ips(port_id=port_id).unique
        self.assertEqual(floating_ip['floating_ip_address'],
                         self.server.floating_ip_address)

    def test_list_floating_ip_with_floating_ip_address(self):
        floating_ip_address = self.server.floating_ip_address
        floating_ip = neutron.list_floating_ips(
            floating_ip_address=floating_ip_address).unique
        self.assertEqual(floating_ip['port_id'],
                         self.server.port_id)

    def test_find_floating_ip_with_port_id(self):
        port_id = self.server.port_id
        floating_ip = neutron.find_floating_ip(port_id=port_id, unique=True)
        self.assertEqual(floating_ip['floating_ip_address'],
                         self.server.floating_ip_address)

    def test_find_floating_ip_with_floating_ip_address(self):
        floating_ip_address = self.server.floating_ip_address
        floating_ip = neutron.find_floating_ip(
            floating_ip_address=floating_ip_address)
        self.assertEqual(floating_ip, self.server.floating_ip_details)

    def test_find_floating_ip_with_invalid_floating_ip_address(self):
        self.assertRaises(tobiko.ObjectNotFound,
                          neutron.find_floating_ip,
                          floating_ip_address='1.2.3.4')



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc
import functools
import inspect
import types
import typing

import decorator


P = typing.TypeVar('P')

GenericMetaBase = abc.ABCMeta
if hasattr(typing, 'GenericMeta'):
    class GenericMetaBase(  # type: ignore[no-redef]
            typing.GenericMeta,  # type: ignore[name-defined]
            abc.ABCMeta):
        # pylint: disable=function-redefined,no-member
        pass


class GenericMeta(GenericMetaBase):

    @functools.lru_cache(maxsize=4096)
    def __getitem__(self, item):
        # pylint: disable=not-callable
        cls = self
        getitem = getattr(super(), '__getitem__', None)
        if callable(getitem):
            cls = getitem(item)
        class_getitem = getattr(cls, '__class_getitem__', None)
        if callable(class_getitem):
            if inspect.ismethod(class_getitem):
                cls = class_getitem(item)
            else:
                cls = class_getitem(cls, item)
        return cls


class Generic(typing.Generic[P], metaclass=GenericMeta):
    pass


PredicateFunction = typing.Callable[[typing.Any], bool]


def is_public_function(obj) -> bool:
    return (inspect.isfunction(obj) and
            getattr(obj, '__name__', '_')[0] != '_')


def is_public_abstract_method(obj) -> bool:
    return (is_public_function(obj) and
            getattr(obj, "__isabstractmethod__", False))


class CallHandler(abc.ABC):

    @abc.abstractmethod
    def _handle_call(self, method: typing.Callable, *args, **kwargs) \
            -> typing.Any:
        raise NotImplementedError


class CallProxyBase(CallHandler, Generic[P], abc.ABC):

    def __class_getitem__(cls, item: typing.Type[P]):
        if isinstance(item, type):
            return create_call_proxy_class(protocols=(item,),
                                           class_name=cls.__name__,
                                           bases=(cls, item),
                                           predicate=cls._is_proxy_method)
        else:
            return cls

    @classmethod
    def _is_proxy_method(cls, obj) -> bool:
        return is_public_abstract_method(obj)


class CallProxy(CallProxyBase, Generic[P]):

    def __init__(self, handle_call: typing.Callable):
        super(CallProxy, self).__init__()
        assert callable(handle_call)
        self._handle_call = handle_call  # type: ignore

    def _handle_call(self, method: typing.Callable, *args, **kwargs):
        raise NotImplementedError


def create_call_proxy_class(
        protocols: typing.Tuple[type, ...],
        class_name: str,
        bases: typing.Tuple[type, ...] = None,
        namespace: dict = None,
        predicate: PredicateFunction = None) -> type:
    if bases is None:
        bases = tuple()
    if predicate is None:
        predicate = is_public_abstract_method

    def exec_body(ns: typing.Dict[str, typing.Any]):
        if namespace is not None:
            ns.update(namespace)
        for cls in protocols:
            for member_name, member in inspect.getmembers(cls, predicate):
                if member_name not in ns:
                    method = create_call_proxy_method(member)
                    ns[member_name] = method

    return types.new_class(name=class_name,
                           bases=bases,
                           exec_body=exec_body)


def create_call_proxy(handle_call: typing.Callable,
                      *protocols: typing.Type[P]) -> P:
    cls = create_call_proxy_class(protocols=protocols,
                                  class_name='CallProxy',
                                  bases=(CallProxy,) + protocols)
    return cls(handle_call)  # type: ignore[call-arg]


def list_abstract_classes(cls: typing.Type) \
        -> typing.Tuple[typing.Type[P], ...]:
    subclasses = inspect.getmro(cls)
    protocols = tuple(cls
                      for cls in subclasses
                      if inspect.isabstract(cls))
    return typing.cast(typing.Tuple[typing.Type[P], ...], protocols)


def list_abstract_methods(cls: typing.Type) \
        -> typing.List[typing.Tuple[str, typing.Callable]]:
    methods: typing.List[typing.Tuple[str, typing.Callable]] = []
    for name, member in inspect.getmembers(cls, inspect.isfunction):
        if getattr(member, "__isabstractmethod__", False):
            methods.append((name, member))
    return methods


def create_call_proxy_method(func: typing.Callable) -> typing.Callable:
    method = decorator.decorate(func, _call_proxy_method)
    assert method is not func
    setattr(method, "__isabstractmethod__", False)
    return method


def _call_proxy_method(func, self: CallProxy, *args, **kwargs):
    # pylint: disable=protected-access
    return self._handle_call(func, *args, **kwargs)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import glance
from tobiko.openstack import stacks


class GlanceApiTestCase(testtools.TestCase):
    """Tests glance images API"""

    #: Stack of resources with a network with a gateway router
    fixture = tobiko.required_fixture(stacks.CirrosImageFixture)

    def test_get_image(self):
        image = glance.get_image(self.fixture.image_id)
        self.assertEqual(self.fixture.image_id, image['id'])

    def test_find_image_by_id(self):
        image = glance.find_image(id=self.fixture.image_id)
        self.assertEqual(self.fixture.image_id, image['id'])

    def test_find_image_by_name(self):
        image = glance.find_image(name=self.fixture.image_name)
        self.assertEqual(self.fixture.image_name, image['name'])



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

IniFileTextType = typing.Union[str, typing.Iterable[str]]


def parse_ini_file(text: IniFileTextType,
                   section='DEFAULT') \
        -> typing.Dict[typing.Tuple[str, str], typing.Any]:
    if isinstance(text, str):
        text = text.splitlines()
    content: typing.Dict[typing.Tuple[str, str], typing.Any] = {}
    for line in text:
        line = line.rsplit('#', 1)[0]  # Remove comments
        line = line.strip()  # strip whitespaces
        if not line:
            # skip empty line
            continue
        if line.startswith('['):
            # parse section name
            section = line[1:-1].strip()
            continue
        if '=' in line:
            left, right = line.split('=', 1)
            content[section, left.strip()] = right.strip()
    return content



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from designateclient.v2 import client

import tobiko
from tobiko.openstack import _client


DESIGNATE_CLIENT_CLASSES = client.Client,


class DesignateClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session) -> client.Client:
        return client.Client(session=session)


class DesignateClientManager(_client.OpenstackClientManager):

    def create_client(self, session) -> DesignateClientFixture:
        return DesignateClientFixture(session=session)


CLIENTS = DesignateClientManager()

DesignateClientType = typing.Union[client.Client, DesignateClientFixture]


def designate_client(obj: DesignateClientType = None):
    if obj is None:
        return get_designate_client()

    if isinstance(obj, client.Client):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, DesignateClientFixture):
        return fixture.client

    message = "Object {!r} is not an OctaviaClientFixture".format(obj)
    raise TypeError(message)


def get_designate_client(session=None,
                         shared=True,
                         init_client=None,
                         manager: DesignateClientManager = None) \
        -> client.Client:
    manager = manager or CLIENTS
    fixture = manager.get_client(session=session,
                                 shared=shared,
                                 init_client=init_client)
    tobiko.setup_fixture(fixture)
    return fixture.client
